from typing import (
    Any,
    Dict,
    List,
    Optional,
    Tuple,
    cast,
)

from eth_rlp import (
    HashableRLP,
)
from eth_utils import (
    ValidationError,
    keccak,
)
from eth_utils.curried import (
    apply_formatters_to_dict,
)
from eth_utils.toolz import (
    dissoc,
    merge,
    partial,
    pipe,
)
from hexbytes import (
    HexBytes,
)
import rlp
from rlp.sedes import (
    Binary,
    CountableList,
    big_endian_int,
    binary,
)

from eth_account._utils.transaction_utils import (
    set_transaction_type_if_needed,
    transaction_rlp_to_rpc_structure,
    transaction_rpc_to_rlp_structure,
)
from eth_account._utils.validation import (
    LEGACY_TRANSACTION_VALID_VALUES,
    is_int_or_prefixed_hexstr,
    is_rpc_structured_access_list,
    is_sequence_of_bytes_or_hexstr,
)
from eth_account.typed_transactions.access_list_transaction import (
    access_list_sede_type,
)
from eth_account.typed_transactions.base import (
    TYPED_TRANSACTION_FORMATTERS,
    Blob,
    BlobPooledTransactionData,
    _TypedTransactionImplementation,
)
from eth_account.types import (
    Blobs,
)


class BlobTransaction(_TypedTransactionImplementation):
    """
    Represents a blob transaction as per EIP-4844.
    """

    transaction_type = 3  # '0x03'
    blob_data: Optional[BlobPooledTransactionData] = None

    unsigned_transaction_fields = (
        ("chainId", big_endian_int),
        ("nonce", big_endian_int),
        ("maxPriorityFeePerGas", big_endian_int),
        ("maxFeePerGas", big_endian_int),
        ("gas", big_endian_int),
        ("to", Binary.fixed_length(20, allow_empty=True)),
        ("value", big_endian_int),
        ("data", binary),
        ("accessList", access_list_sede_type),
        ("maxFeePerBlobGas", big_endian_int),
        (
            "blobVersionedHashes",
            CountableList(Binary.fixed_length(32, allow_empty=False)),
        ),
    )

    signature_fields = (
        ("v", big_endian_int),
        ("r", big_endian_int),
        ("s", big_endian_int),
    )

    transaction_field_defaults = {
        "type": b"0x3",
        "chainId": 0,
        "to": b"",
        "value": 0,
        "data": b"",
        "accessList": [],
    }

    _unsigned_transaction_serializer = type(
        "_unsigned_transaction_serializer",
        (HashableRLP,),
        {
            "fields": unsigned_transaction_fields,
        },
    )

    _signed_transaction_serializer = type(
        "_signed_transaction_serializer",
        (HashableRLP,),
        {
            "fields": unsigned_transaction_fields + signature_fields,
        },
    )

    _signed_pooled_transaction_serializer = type(
        "_signed_pooled_transaction_serializer",
        (HashableRLP,),
        {
            "fields": (
                ("tx_payload_body", _signed_transaction_serializer),
                (
                    "blobs",
                    CountableList(binary.fixed_length(4096 * 32)),
                ),
                (
                    "commitments",
                    CountableList(binary.fixed_length(48)),
                ),
                ("proofs", CountableList(binary.fixed_length(48))),
            ),
        },
    )

    def __init__(
        self,
        dictionary: Dict[str, Any],
        blobs: Optional[Blobs] = None,
    ):
        self.dictionary = dictionary

        if blobs is not None:
            self.blob_data = BlobPooledTransactionData(
                blobs=[Blob(data=HexBytes(blob_data)) for blob_data in blobs]
            )
            if "blobVersionedHashes" in dictionary:
                self._validate_versioned_hashes_against_blob_data(
                    dictionary["blobVersionedHashes"],
                    self.blob_data,
                )

    @classmethod
    def assert_valid_fields(
        cls,
        dictionary: Dict[str, Any],
        has_blobs: bool = False,
    ) -> None:
        transaction_valid_values = merge(
            LEGACY_TRANSACTION_VALID_VALUES,
            {
                "type": is_int_or_prefixed_hexstr,
                "maxPriorityFeePerGas": is_int_or_prefixed_hexstr,
                "maxFeePerGas": is_int_or_prefixed_hexstr,
                "accessList": is_rpc_structured_access_list,
                "maxFeePerBlobGas": is_int_or_prefixed_hexstr,
            },
        )
        if not has_blobs:
            transaction_valid_values[
                "blobVersionedHashes"
            ] = is_sequence_of_bytes_or_hexstr(item_bytes_size=32, can_be_empty=False)

        if "v" in dictionary and dictionary["v"] == 0:
            dictionary["v"] = "0x0"
        valid_fields = apply_formatters_to_dict(
            transaction_valid_values,
            dictionary,
        )  # type: Dict[str, Any]
        if not all(valid_fields.values()):
            invalid = {
                key: dictionary[key] for key, valid in valid_fields.items() if not valid
            }
            raise TypeError(f"Transaction had invalid fields: {repr(invalid)}")

    @classmethod
    def from_dict(
        cls,
        dictionary: Dict[str, Any],
        blobs: Optional[Blobs] = None,
    ) -> "BlobTransaction":
        """
        Builds a BlobTransaction from a dictionary.
        Verifies that the dictionary is well-formed.
        """
        has_blobs = blobs is not None

        if "tx_payload_body" in dictionary:
            dictionary = dict(
                zip(
                    (
                        entry[0]
                        for entry in cls.unsigned_transaction_fields
                        + cls.signature_fields
                    ),
                    dictionary["tx_payload_body"],
                )
            )
            dictionary["type"] = cls.transaction_type
        else:
            dictionary = set_transaction_type_if_needed(dictionary)

        # Validate fields.
        cls.assert_valid_fields(dictionary, has_blobs=has_blobs)
        sanitized_dictionary = pipe(
            dictionary,
            dict,
            partial(merge, cls.transaction_field_defaults),
            apply_formatters_to_dict(TYPED_TRANSACTION_FORMATTERS),
        )

        # We have verified the type, we can safely remove it from the dictionary,
        # given that it is not to be included within the RLP payload.
        transaction_type = sanitized_dictionary.pop("type")
        if transaction_type != cls.transaction_type:
            raise ValueError(
                f"expected transaction type {cls.transaction_type}, "
                f"got {transaction_type}"
            )
        return cls(
            dictionary=sanitized_dictionary,
            blobs=blobs,
        )

    @classmethod
    def from_bytes(cls, encoded_transaction: HexBytes) -> "BlobTransaction":
        """
        Builds a BlobTransaction from a signed encoded transaction.
        """
        if not isinstance(encoded_transaction, HexBytes):
            raise TypeError(f"expected Hexbytes, got type: {type(encoded_transaction)}")
        if not (
            len(encoded_transaction) > 0
            and encoded_transaction[0] == cls.transaction_type
        ):
            raise ValueError("unexpected input")

        # Format is (0x03 || TransactionPayload)
        # We strip the prefix, and RLP unmarshal the payload into our
        # signed transaction serializer.
        transaction_payload = encoded_transaction[1:]
        try:
            # Attempt to deserialize as a `PooledTransaction`, as defined in EIP-4844.
            dictionary = cls._signed_pooled_transaction_serializer.from_bytes(  # type: ignore  # noqa: E501
                transaction_payload
            ).as_dict()
        except rlp.exceptions.ObjectDeserializationError:
            # If the deserialization fails, we attempt to deserialize as a
            # `TransactionPayloadBody`, as defined in EIP-4844.
            dictionary = cls._signed_transaction_serializer.from_bytes(  # type: ignore  # noqa: E501
                transaction_payload
            ).as_dict()

        rpc_structured_dict = transaction_rlp_to_rpc_structure(dictionary)
        rpc_structured_dict["type"] = cls.transaction_type
        blobs = dictionary.get("blobs")
        return cls.from_dict(rpc_structured_dict, blobs=blobs)

    def as_dict(self) -> Dict[str, Any]:
        """Returns this transaction as a dictionary."""
        dictionary = self.dictionary.copy()
        dictionary["type"] = self.__class__.transaction_type

        if self.blob_data is not None:
            if "blobVersionedHashes" in dictionary:
                # we validate our versioned hashes calculation internally so this is a
                # safer way to ensure we're returning correct data.
                self._validate_versioned_hashes_against_blob_data(
                    dictionary["blobVersionedHashes"], self.blob_data
                )
            else:
                # If versioned hashes are not provided, we compute them.
                dictionary["blobVersionedHashes"] = [
                    versioned_hash.data
                    for versioned_hash in self.blob_data.versioned_hashes
                ]

        return dictionary

    def hash(self) -> bytes:
        """
        Keccak256 hash of the BlobTransaction to prepare it for signing.
        As per the EIP-4844 specifications, the signature is a secp256k1 signature over
        ``keccak256(0x03 || rlp([chainId, nonce, maxPriorityFeePerGas,
        maxFeePerGas, gasLimit, to, value, data, accessList, maxFeePerBlobGas,
        blobVersionedHashes]))``.
        """
        # Remove signature fields.
        transaction_without_signature_fields = dissoc(self.dictionary, "v", "r", "s")
        # RPC-structured transaction to rlp-structured transaction
        rlp_structured_txn_without_sig_fields = transaction_rpc_to_rlp_structure(
            transaction_without_signature_fields
        )

        if self.blob_data is not None:
            if rlp_structured_txn_without_sig_fields.get("blobVersionedHashes") is None:
                # If the versioned hashes are not provided, we compute them.
                rlp_structured_txn_without_sig_fields["blobVersionedHashes"] = [
                    versioned_hash.data
                    for versioned_hash in self.blob_data.versioned_hashes
                ]
            else:
                # Validate that the versioned hashes match the computed versioned
                # hashes.
                self._validate_versioned_hashes_against_blob_data(
                    rlp_structured_txn_without_sig_fields["blobVersionedHashes"],
                    self.blob_data,
                )

        rlp_serializer = self.__class__._unsigned_transaction_serializer
        hash_ = pipe(
            rlp_serializer.from_dict(rlp_structured_txn_without_sig_fields),  # type: ignore  # noqa: E501
            lambda val: rlp.encode(val),  # rlp([...])
            lambda val: bytes([self.__class__.transaction_type])
            + val,  # (0x03 || rlp([...]))
            keccak,  # keccak256(0x03 || rlp([...]))
        )
        return cast(bytes, hash_)

    def payload(self) -> bytes:
        """
        Returns this transaction's payload as bytes.

        Here, the transaction payload is:

            TransactionPayload = rlp([chainId,
            nonce, maxPriorityFeePerGas, maxFeePerGas, gasLimit, to, value, data,
            accessList, maxFeePerBlobGas, blobVersionedHashes, signatureYParity,
            signatureR, signatureS])

        """
        if not all(k in self.dictionary for k in "vrs"):
            raise ValueError("attempting to encode an unsigned transaction")

        rlp_structured_dict = transaction_rpc_to_rlp_structure(self.dictionary)
        if self.blob_data is None:
            # `TransactionPayload` as defined in EIP-4844
            # rlp([tx_payload_body])
            rlp_serializer = self.__class__._signed_transaction_serializer
            payload = rlp.encode(rlp_serializer.from_dict(rlp_structured_dict))  # type: ignore # noqa: E501
        else:
            # `PooledTransaction` as defined in EIP-4844
            # rlp([tx_payload_body, blobs, commitments, proofs])
            rlp_serializer = self.__class__._signed_pooled_transaction_serializer
            pooled_txn_as_dict = {
                "tx_payload_body": tuple(
                    rlp_structured_dict[key]
                    for key, _val in (
                        self.unsigned_transaction_fields + self.signature_fields
                    )
                ),
                "blobs": [blob.as_bytes() for blob in self.blob_data.blobs],
                "commitments": [
                    commitment.as_bytes() for commitment in self.blob_data.commitments
                ],
                "proofs": [proof.as_bytes() for proof in self.blob_data.proofs],
            }
            payload = rlp.encode(rlp_serializer.from_dict(pooled_txn_as_dict))  # type: ignore # noqa: E501

        return cast(bytes, payload)

    def vrs(self) -> Tuple[int, int, int]:
        """Returns (v, r, s) if they exist."""
        if not all(k in self.dictionary for k in "vrs"):
            raise ValueError("attempting to encode an unsigned transaction")
        return (self.dictionary["v"], self.dictionary["r"], self.dictionary["s"])

    @staticmethod
    def _validate_versioned_hashes_against_blob_data(
        blob_versioned_hashes: List[bytes],
        blob_data: BlobPooledTransactionData,
    ) -> None:
        diff = set(blob_versioned_hashes).difference(
            {versioned_hash.data for versioned_hash in blob_data.versioned_hashes}
        )
        if diff:
            raise ValidationError(
                "`blobVersionedHashes` value defined in transaction does not match "
                f"versioned hashes computed from blobs.\n    diff: {diff}"
            )



from typing import (
    Any,
    Dict,
    NamedTuple,
    Optional,
    Union,
)

from eth_typing import (
    Address,
    Hash32,
)
from eth_utils.curried import (
    ValidationError,
    keccak,
    text_if_str,
    to_bytes,
    to_canonical_address,
)
from hexbytes import (
    HexBytes,
)

from eth_account._utils.encode_typed_data.encoding_and_hashing import (
    get_primary_type,
    hash_domain,
    hash_eip712_message,
)
from eth_account._utils.validation import (
    is_valid_address,
)

text_to_bytes = text_if_str(to_bytes)


# watch for updates to signature format
class SignableMessage(NamedTuple):
    """
    A message compatible with EIP-191_ that is ready to be signed.

    The properties are components of an EIP-191_ signable message. Other message formats
    can be encoded into this format for easy signing. This data structure doesn't need
    to know about the original message format. For example, you can think of
    EIP-712 as compiling down to an EIP-191 message.

    In typical usage, you should never need to create these by hand. Instead, use
    one of the available encode_* methods in this module, like:

        - :meth:`encode_intended_validator`
        - :meth:`encode_defunct`
        - :meth:`encode_typed_data`

    .. _EIP-191: https://eips.ethereum.org/EIPS/eip-191
    """

    version: bytes  # must be length 1
    header: bytes  # aka "version specific data"
    body: bytes  # aka "data to sign"


def _hash_eip191_message(signable_message: SignableMessage) -> Hash32:
    version = signable_message.version
    if len(version) != 1:
        raise ValidationError(
            f"The supplied message version is {version!r}. "
            "The EIP-191 signable message standard only supports one-byte versions."
        )

    joined = b"\x19" + version + signable_message.header + signable_message.body
    return Hash32(keccak(joined))


# watch for updates to signature format
def encode_intended_validator(
    validator_address: Union[Address, str],
    primitive: Optional[bytes] = None,
    *,
    hexstr: Optional[str] = None,
    text: Optional[str] = None,
) -> SignableMessage:
    """
    Encode a message using the "intended validator" approach (ie~ version 0)
    defined in EIP-191_.

    Supply the message as exactly one of these three arguments:
    bytes as a primitive, a hex string, or a unicode string.

    .. WARNING:: Note that this code has not gone through an external audit.

    :param validator_address: which on-chain contract is capable of validating this
        message, provided as a checksummed address or in native bytes.
    :param primitive: the binary message to be signed
    :type primitive: bytes or int
    :param str hexstr: the message encoded as hex
    :param str text: the message as a series of unicode characters (a normal Py3 str)
    :returns: The EIP-191 encoded message, ready for signing

    .. _EIP-191: https://eips.ethereum.org/EIPS/eip-191
    """
    if not is_valid_address(validator_address):
        raise ValidationError(
            f"Cannot encode message with 'Validator Address': {validator_address!r}. "
            "It must be a checksum address, or an address converted to bytes."
        )
    # The validator_address is a str or Address (which is a subtype of bytes). Both of
    # these are AnyStr, which includes str and bytes.
    # Not sure why mypy complains here...
    canonical_address = to_canonical_address(validator_address)
    message_bytes = to_bytes(primitive, hexstr=hexstr, text=text)
    return SignableMessage(
        HexBytes(b"\x00"),  # version 0, as defined in EIP-191
        canonical_address,
        message_bytes,
    )


def encode_defunct(
    primitive: Optional[bytes] = None,
    *,
    hexstr: Optional[str] = None,
    text: Optional[str] = None,
) -> SignableMessage:
    r"""
    Encode a message for signing, using an old, unrecommended approach.

    Only use this method if you must have compatibility with
    :meth:`w3.eth.sign() <web3.eth.Eth.sign>`.

    EIP-191 defines this as "version ``E``".

    .. NOTE: This standard includes the number of bytes in the message as a part of
        the header. Awkwardly, the number of bytes in the message is encoded in
        decimal ascii. So if the message is 'abcde', then the length is encoded
        as the ascii character '5'. This is one of the reasons that this message
        format is not preferred. There is ambiguity when the message '00' is
        encoded, for example.

    Supply exactly one of the three arguments: bytes, a hex string, or a unicode string.

    :param primitive: the binary message to be signed
    :type primitive: bytes or int
    :param str hexstr: the message encoded as hex
    :param str text: the message as a series of unicode characters (a normal Py3 str)
    :returns: The EIP-191 encoded message, ready for signing

    .. doctest:: python

        >>> from eth_account.messages import encode_defunct
        >>> from eth_utils.curried import to_hex, to_bytes

        >>> message_text = "I♥SF"
        >>> encode_defunct(text=message_text)
        SignableMessage(version=b'E',
                        header=b'thereum Signed Message:\n6',
                        body=b'I\xe2\x99\xa5SF')

        These four also produce the same hash:
        >>> encode_defunct(to_bytes(text=message_text))
        SignableMessage(version=b'E',
                        header=b'thereum Signed Message:\n6',
                        body=b'I\xe2\x99\xa5SF')

        >>> encode_defunct(bytes(message_text, encoding='utf-8'))
        SignableMessage(version=b'E',
                        header=b'thereum Signed Message:\n6',
                        body=b'I\xe2\x99\xa5SF')

        >>> to_hex(text=message_text)
        '0x49e299a55346'
        >>> encode_defunct(hexstr='0x49e299a55346')
        SignableMessage(version=b'E',
                        header=b'thereum Signed Message:\n6',
                        body=b'I\xe2\x99\xa5SF')

        >>> encode_defunct(0x49e299a55346)
        SignableMessage(version=b'E',
                        header=b'thereum Signed Message:\n6',
                        body=b'I\xe2\x99\xa5SF')
    """
    message_bytes = to_bytes(primitive, hexstr=hexstr, text=text)
    msg_length = str(len(message_bytes)).encode("utf-8")

    # Encoding version E defined by EIP-191
    return SignableMessage(
        b"E",
        b"thereum Signed Message:\n" + msg_length,
        message_bytes,
    )


def defunct_hash_message(
    primitive: Optional[bytes] = None,
    *,
    hexstr: Optional[str] = None,
    text: Optional[str] = None,
) -> HexBytes:
    """
    Convert the provided message into a message hash, to be signed.

    .. CAUTION:: Intended for use with
        :meth:`eth_account.account.Account.unsafe_sign_hash`.
        This is for backwards compatibility only. All new implementations
        should use :meth:`encode_defunct` instead.

    :param primitive: the binary message to be signed
    :type primitive: bytes or int
    :param str hexstr: the message encoded as hex
    :param str text: the message as a series of unicode characters (a normal Py3 str)
    :returns: The hash of the message, after adding the prefix
    """
    signable = encode_defunct(primitive, hexstr=hexstr, text=text)
    hashed = _hash_eip191_message(signable)
    return HexBytes(hashed)


def encode_typed_data(
    domain_data: Optional[Dict[str, Any]] = None,
    message_types: Optional[Dict[str, Any]] = None,
    message_data: Optional[Dict[str, Any]] = None,
    full_message: Optional[Dict[str, Any]] = None,
) -> SignableMessage:
    r"""
    Encode an EIP-712_ message in a manner compatible with other implementations
    in use, such as the Metamask and Ethers ``signTypedData`` functions.

    See the `EIP-712 spec <https://eips.ethereum.org/EIPS/eip-712>`_ for more information.

    You may supply the information to be encoded in one of two ways:

    As exactly three arguments:

        - ``domain_data``, a dict of the EIP-712 domain data
        - ``message_types``, a dict of custom types (do not include a ``EIP712Domain``
          key)
        - ``message_data``, a dict of the data to be signed

    Or as a single argument:

        - ``full_message``, a dict containing the following keys:
            - ``types``, a dict of custom types (may include a ``EIP712Domain`` key)
            - ``primaryType``, (optional) a string of the primary type of the message
            - ``domain``, a dict of the EIP-712 domain data
            - ``message``, a dict of the data to be signed

    .. WARNING:: Note that this code has not gone through an external audit, and
        the test cases are incomplete.

    Type Coercion:
        - For fixed-size bytes types, smaller values will be padded to fit in larger
          types, but values larger than the type will raise ``ValueOutOfBounds``.
          e.g., an 8-byte value will be padded to fit a ``bytes16`` type, but 16-byte
          value provided for a ``bytes8`` type will raise an error.
        - Fixed-size and dynamic ``bytes`` types will accept ``int``s. Any negative
          values will be converted to ``0`` before being converted to ``bytes``
        - ``int`` and ``uint`` types will also accept strings. If prefixed with ``"0x"``
          , the string will be interpreted as hex. Otherwise, it will be interpreted as
          decimal.

    Noteable differences from ``signTypedData``:
        - Custom types that are not alphanumeric will encode differently.
        - Custom types that are used but not defined in ``types`` will not encode.

    :param domain_data: EIP712 domain data
    :param message_types: custom types used by the `value` data
    :param message_data: data to be signed
    :param full_message: a dict containing all data and types
    :returns: a ``SignableMessage``, an encoded message ready to be signed


    .. doctest:: python

        >>> # examples of basic usage
        >>> from eth_account import Account
        >>> from eth_account.messages import encode_typed_data
        >>> # 3-argument usage

        >>> # all domain properties are optional
        >>> domain_data = {
        ...     "name": "Ether Mail",
        ...     "version": "1",
        ...     "chainId": 1,
        ...     "verifyingContract": "0xCcCCccccCCCCcCCCCCCcCcCccCcCCCcCcccccccC",
        ...     "salt": b"decafbeef",
        ... }
        >>> # custom types
        >>> message_types = {
        ...     "Person": [
        ...         {"name": "name", "type": "string"},
        ...         {"name": "wallet", "type": "address"},
        ...     ],
        ...     "Mail": [
        ...         {"name": "from", "type": "Person"},
        ...         {"name": "to", "type": "Person"},
        ...         {"name": "contents", "type": "string"},
        ...     ],
        ... }
        >>> # the data to be signed
        >>> message_data = {
        ...     "from": {
        ...         "name": "Cow",
        ...         "wallet": "0xCD2a3d9F938E13CD947Ec05AbC7FE734Df8DD826",
        ...     },
        ...     "to": {
        ...         "name": "Bob",
        ...         "wallet": "0xbBbBBBBbbBBBbbbBbbBbbbbBBbBbbbbBbBbbBBbB",
        ...     },
        ...     "contents": "Hello, Bob!",
        ... }
        >>> key = "0xaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
        >>> signable_message = encode_typed_data(domain_data, message_types, message_data)
        >>> signed_message = Account.sign_message(signable_message, key)
        >>> signed_message.message_hash
        HexBytes('0xc5bb16ccc59ae9a3ad1cb8343d4e3351f057c994a97656e1aff8c134e56f7530')
        >>> # the message can be signed in one step using Account.sign_typed_data
        >>> signed_typed_data = Account.sign_typed_data(key, domain_data, message_types, message_data)
        >>> signed_typed_data == signed_message
        True

        >>> # 1-argument usage

        >>> # all domain properties are optional
        >>> full_message = {
        ...     "types": {
        ...         "EIP712Domain": [
        ...             {"name": "name", "type": "string"},
        ...             {"name": "version", "type": "string"},
        ...             {"name": "chainId", "type": "uint256"},
        ...             {"name": "verifyingContract", "type": "address"},
        ...             {"name": "salt", "type": "bytes32"},
        ...         ],
        ...         "Person": [
        ...             {"name": "name", "type": "string"},
        ...             {"name": "wallet", "type": "address"},
        ...         ],
        ...         "Mail": [
        ...             {"name": "from", "type": "Person"},
        ...             {"name": "to", "type": "Person"},
        ...             {"name": "contents", "type": "string"},
        ...         ],
        ...     },
        ...     "primaryType": "Mail",
        ...     "domain": {
        ...         "name": "Ether Mail",
        ...         "version": "1",
        ...         "chainId": 1,
        ...         "verifyingContract": "0xCcCCccccCCCCcCCCCCCcCcCccCcCCCcCcccccccC",
        ...         "salt": b"decafbeef"
        ...     },
        ...     "message": {
        ...         "from": {
        ...             "name": "Cow",
        ...             "wallet": "0xCD2a3d9F938E13CD947Ec05AbC7FE734Df8DD826"
        ...         },
        ...         "to": {
        ...             "name": "Bob",
        ...             "wallet": "0xbBbBBBBbbBBBbbbBbbBbbbbBBbBbbbbBbBbbBBbB"
        ...         },
        ...         "contents": "Hello, Bob!",
        ...     },
        ... }
        >>> signable_message_2 = encode_typed_data(full_message=full_message)
        >>> signed_message_2 = Account.sign_message(signable_message_2, key)
        >>> signed_message_2.message_hash
        HexBytes('0xc5bb16ccc59ae9a3ad1cb8343d4e3351f057c994a97656e1aff8c134e56f7530')
        >>> signed_message_2 == signed_message
        True
        >>> # the full_message can be signed in one step using Account.sign_typed_data
        >>> signed_typed_data_2 = Account.sign_typed_data(key, domain_data, message_types, message_data)
        >>> signed_typed_data_2 == signed_message_2
        True

    .. _EIP-712: https://eips.ethereum.org/EIPS/eip-712
    """  # noqa: E501
    if full_message is not None:
        if (
            domain_data is not None
            or message_types is not None
            or message_data is not None
        ):
            raise ValueError(
                "You may supply either `full_message` as a single argument or "
                "`domain_data`, `message_types`, and `message_data` as three arguments,"
                " but not both."
            )

        full_message_types = full_message["types"].copy()
        full_message_domain = full_message["domain"].copy()

        # If EIP712Domain types were provided, check that they match the domain data
        if "EIP712Domain" in full_message_types:
            domain_data_keys = list(full_message_domain.keys())
            domain_types_keys = [
                field["name"] for field in full_message_types["EIP712Domain"]
            ]

            if set(domain_data_keys) != (set(domain_types_keys)):
                raise ValidationError(
                    "The fields provided in `domain` do not match the fields provided"
                    " in `types.EIP712Domain`. The fields provided in `domain` were"
                    f" `{domain_data_keys}`, but the fields provided in "
                    f"`types.EIP712Domain` were `{domain_types_keys}`."
                )

        full_message_types.pop("EIP712Domain", None)

        # If primaryType was provided, check that it matches the derived primaryType
        if "primaryType" in full_message:
            derived_primary_type = get_primary_type(full_message_types)
            provided_primary_type = full_message["primaryType"]
            if derived_primary_type != provided_primary_type:
                raise ValidationError(
                    "The provided `primaryType` does not match the derived "
                    "`primaryType`. The provided `primaryType` was "
                    f"`{provided_primary_type}`, but the derived `primaryType` was "
                    f"`{derived_primary_type}`."
                )

        parsed_domain_data = full_message_domain
        parsed_message_types = full_message_types
        parsed_message_data = full_message["message"]

    else:
        parsed_domain_data = domain_data
        parsed_message_types = message_types
        parsed_message_data = message_data

    return SignableMessage(
        HexBytes(b"\x01"),
        hash_domain(parsed_domain_data),
        hash_eip712_message(parsed_message_types, parsed_message_data),
    )



import pytest
import subprocess

from hypothesis import (
    given,
    settings,
    strategies as st,
)

from eth_account import (
    Account,
)
from eth_account.hdaccount.mnemonic import (
    VALID_ENTROPY_SIZES,
    Mnemonic,
)

Account.enable_unaudited_hdwallet_features()

language_st = st.sampled_from(Mnemonic.list_languages())

seed_st = (
    st.binary(min_size=min(VALID_ENTROPY_SIZES), max_size=max(VALID_ENTROPY_SIZES))
    .filter(lambda x: len(x) in VALID_ENTROPY_SIZES)
    .filter(lambda s: int.from_bytes(s, byteorder="big") != 0)
)

node_st = st.tuples(st.integers(min_value=0, max_value=2**31 - 1), st.booleans())
path_st = (
    st.lists(node_st, min_size=0, max_size=10)
    .map(lambda nodes: list(str(n[0]) + ("" if n[1] else "'") for n in nodes))
    .map(lambda nodes: "m" + ("/" + "/".join(nodes) if nodes else ""))
)


@given(seed=seed_st, language=language_st, account_path=path_st)
@settings(deadline=1000)
@pytest.mark.compatibility
def test_compatibility(seed, language, account_path):
    mnemonic = Mnemonic(language).to_mnemonic(seed)
    acct = Account.from_mnemonic(mnemonic, account_path=account_path)
    # NOTE Must do `cd tests/integration/js-scripts && npm install -g .
    ethers_cli = subprocess.run(
        [
            "node",
            "tests/integration/js-scripts/ethers-mnemonic-fuzzing",
            "-m",
            mnemonic,
            "-l",
            language,
            "-p",
            account_path,
        ],
        capture_output=True,
    )
    if ethers_cli.stderr:
        raise OSError(ethers_cli.stderr.decode("utf-8"))
    ethers_address = ethers_cli.stdout.decode("utf-8").strip()
    assert acct.address == ethers_address



# Originally from: https://github.com/trezor/python-mnemonic
#
# Copyright (c) 2013 Pavol Rusnak
# Copyright (c) 2017 mruddy
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

import os
from pathlib import (
    Path,
)
import secrets
from typing import (
    Dict,
    List,
)

from bitarray import (
    bitarray,
)
from bitarray.util import (
    ba2int,
    int2ba,
)
from eth_utils import (
    ValidationError,
)

from ._utils import (
    normalize_string,
    pbkdf2_hmac_sha512,
    sha256,
)

VALID_ENTROPY_SIZES = [16, 20, 24, 28, 32]
VALID_WORD_COUNTS = [12, 15, 18, 21, 24]
WORDLIST_DIR = Path(__file__).parent / "wordlist"
WORDLIST_LEN = 2048

_cached_wordlists: Dict[str, List[str]] = dict()


def get_wordlist(language: str) -> List[str]:
    if language in _cached_wordlists.keys():
        return _cached_wordlists[language]
    with open(WORDLIST_DIR / f"{language}.txt", encoding="utf-8") as f:
        wordlist = [w.strip() for w in f.readlines()]
    if len(wordlist) != WORDLIST_LEN:
        raise ValidationError(
            f"Wordlist should contain {WORDLIST_LEN} words, "
            f"but it contains {len(wordlist)} words."
        )
    _cached_wordlists[language] = wordlist
    return wordlist


class Mnemonic:
    def __init__(self, raw_language: str = "english"):
        language = raw_language.lower().replace(" ", "_")
        languages = Mnemonic.list_languages()
        if language not in languages:
            raise ValidationError(
                f'Invalid language choice "{language}", must be one of {languages}'
            )
        self.language = language
        self.wordlist = get_wordlist(language)

    @staticmethod
    def list_languages() -> List[str]:
        return sorted(Path(f).stem for f in WORDLIST_DIR.rglob("*.txt"))

    @classmethod
    def detect_language(cls, raw_mnemonic: str) -> str:
        mnemonic = normalize_string(raw_mnemonic)

        words = set(mnemonic.split(" "))
        matching_languages = {
            lang
            for lang in Mnemonic.list_languages()
            if len(words.intersection(cls(lang).wordlist)) == len(words)
        }

        # No language had all words match it, so the language can't be fully determined
        if len(matching_languages) < 1:
            raise ValidationError(f"Language not detected for word(s): {raw_mnemonic}")

        # If both chinese simplified and chinese traditional match (because one is a
        # subset of the other) then return simplified. This doesn't hold for
        # other languages.
        if len(matching_languages) == 2 and all(
            "chinese" in lang for lang in matching_languages
        ):
            return "chinese_simplified"

        # Because certain wordlists share some similar words, if we detect multiple
        # languages that the provided mnemonic word(s) could be valid in, we have
        # to throw
        if len(matching_languages) > 1:
            raise ValidationError(
                f"Word(s) are valid in multiple languages: {raw_mnemonic}"
            )

        (language,) = matching_languages
        return language

    def generate(self, num_words: int = 12) -> str:
        if num_words not in VALID_WORD_COUNTS:
            raise ValidationError(
                f"Invalid choice for number of words: {num_words}, should be one of "
                f"{VALID_WORD_COUNTS}"
            )
        return self.to_mnemonic(os.urandom(4 * num_words // 3))  # 4/3 bytes per word

    def to_mnemonic(self, entropy: bytes) -> str:
        entropy_size = len(entropy)
        if entropy_size not in VALID_ENTROPY_SIZES:
            raise ValidationError(
                f"Invalid data length {len(entropy)}, should be one of "
                f"{VALID_ENTROPY_SIZES}"
            )

        bits = bitarray()
        bits.frombytes(entropy)

        checksum = bitarray()
        checksum.frombytes(sha256(entropy))

        # Add enough bits from the checksum to make it modulo 11 (2**11 = 2048)
        bits.extend(checksum[: entropy_size // 4])
        indices = tuple(
            ba2int(bits[i * 11 : (i + 1) * 11]) for i in range(len(bits) // 11)
        )
        words = tuple(self.wordlist[idx] for idx in indices)

        if self.language == "japanese":  # Japanese must be joined by ideographic space.
            phrase = "\u3000".join(words)
        else:
            phrase = " ".join(words)
        return phrase

    def is_mnemonic_valid(self, mnemonic: str) -> bool:
        words = normalize_string(mnemonic).split(" ")
        num_words = len(words)

        if num_words not in VALID_WORD_COUNTS:
            return False

        try:
            indices = tuple(self.wordlist.index(w) for w in words)
        except ValueError:
            return False

        encoded_seed = bitarray()
        for idx in indices:
            # Build bitarray from tightly packing indices (which are 11-bits integers)
            encoded_seed.extend(int2ba(idx, length=11))

        entropy_size = 4 * num_words // 3

        # Checksum the raw entropy bits
        checksum = bitarray()
        checksum.frombytes(sha256(encoded_seed[: entropy_size * 8].tobytes()))
        computed_checksum = checksum[: len(encoded_seed) - entropy_size * 8].tobytes()

        # Extract the stored checksum bits
        stored_checksum = encoded_seed[entropy_size * 8 :].tobytes()

        # Check that the stored matches the relevant slice of the actual checksum
        # NOTE: Use secrets.compare_digest for protection again timing attacks
        return secrets.compare_digest(stored_checksum, computed_checksum)

    def expand_word(self, prefix: str) -> str:
        if prefix in self.wordlist:
            return prefix
        else:
            matches: List[str] = [
                word for word in self.wordlist if word.startswith(prefix)
            ]
            if len(matches) == 1:  # matched exactly one word in the wordlist
                return matches[0]
            else:
                # exact match not found.
                # this is not a validation routine, just return the input
                return prefix

    def expand(self, mnemonic: str) -> str:
        return " ".join(map(self.expand_word, mnemonic.split(" ")))

    @classmethod
    def to_seed(cls, checked_mnemonic: str, passphrase: str = "") -> bytes:
        """
        :param str checked_mnemonic: Must be a correct, fully-expanded BIP39 seed phrase
        :param str passphrase: Encryption passphrase used to secure the mnemonic
        :returns bytes: 64 bytes of raw seed material from PRNG
        """
        mnemonic = normalize_string(checked_mnemonic)
        # NOTE: This domain separater ("mnemonic") is added per BIP39 spec
        # to the passphrase
        # https://github.com/bitcoin/bips/blob/master/bip-0039.mediawiki#from-mnemonic-to-seed  # blocklint: URL pragma  # noqa: E501
        salt = "mnemonic" + normalize_string(passphrase)
        # From BIP39:
        #   To create a binary seed from the mnemonic, we use the PBKDF2 function with a
        # mnemonic sentence (in UTF-8 NFKD) used as the password and the string
        # "mnemonic" and passphrase (again in UTF-8 NFKD) used as the salt.
        stretched = pbkdf2_hmac_sha512(mnemonic, salt)
        return stretched[:64]



from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    Any,
)

from eth_typing import (
    ChecksumAddress,
    Hash32,
)

from eth_account.datastructures import (
    SignedMessage,
    SignedTransaction,
)
from eth_account.messages import (
    SignableMessage,
)
from eth_account.types import (
    TransactionDictType,
)


class BaseAccount(ABC):
    """
    Specify convenience methods to sign transactions and message hashes.
    """

    @property
    @abstractmethod
    def address(self) -> ChecksumAddress:
        """
        The checksummed public address for this account.

        .. code-block:: python

            >>> my_account.address # doctest: +SKIP
            "0xF0109fC8DF283027b6285cc889F5aA624EaC1F55"

        """

    @abstractmethod
    def sign_message(self, signable_message: SignableMessage) -> SignedMessage:
        """
        Sign the EIP-191_ message.

        This uses the same structure
        as in :meth:`~eth_account.account.Account.sign_message`
        but without specifying the private key.

        :param signable_message: The encoded message, ready for signing

        .. _EIP-191: https://eips.ethereum.org/EIPS/eip-191
        """

    @abstractmethod
    def unsafe_sign_hash(self, message_hash: Hash32) -> SignedMessage:
        """
        Sign the hash of a message.

        .. WARNING:: *Never* sign a hash that you didn't generate,
            it can be an arbitrary transaction. For example, it might
            send all of your account's ether to an attacker.
            Instead, prefer :meth:`~eth_account.account.Account.sign_message`,
            which cannot accidentally sign a transaction.

        This uses the same structure
        as in :meth:`~eth_account.account.Account.unsafe_sign_hash`
        but without specifying the private key.

        :param bytes message_hash: 32 byte hash of the message to sign
        """

    @abstractmethod
    def sign_transaction(
        self, transaction_dict: TransactionDictType
    ) -> SignedTransaction:
        """
        Sign a transaction dict.

        This uses the same structure as in
        :meth:`~eth_account.account.Account.sign_transaction`
        but without specifying the private key.

        :param dict transaction_dict: transaction with all fields specified
        """

    def __eq__(self, other: Any) -> bool:
        """
        Equality test between two accounts.

        Two accounts are considered the same if they are exactly the same type,
        and can sign for the same address.
        """
        return type(self) is type(other) and self.address == other.address

    def __hash__(self) -> int:
        return hash((type(self), self.address))



from typing import (
    Any,
    Dict,
    Optional,
    cast,
)

from eth_keyfile.keyfile import (
    KDFType,
)
from eth_keys.datatypes import (
    PrivateKey,
)
from eth_typing import (
    ChecksumAddress,
    Hash32,
)

from eth_account.account_local_actions import (
    AccountLocalActions,
)
from eth_account.datastructures import (
    SignedMessage,
    SignedTransaction,
)
from eth_account.messages import (
    SignableMessage,
)
from eth_account.signers.base import (
    BaseAccount,
)
from eth_account.types import (
    Blobs,
    TransactionDictType,
)


class LocalAccount(BaseAccount):
    r"""
    A collection of convenience methods to sign and encrypt, with an
    embedded private key.

    :var bytes key: the 32-byte private key data

    .. code-block:: python

        >>> my_local_account.address
        "0xF0109fC8DF283027b6285cc889F5aA624EaC1F55"
        >>> my_local_account.key
        b"\x01\x23..."

    You can also get the private key by casting the account to :class:`bytes`:

    .. code-block:: python

        >>> bytes(my_local_account)
        b"\\x01\\x23..."
    """

    def __init__(self, key: PrivateKey, account: AccountLocalActions):
        """
        Initialize a new account with the given private key.

        :param eth_keys.PrivateKey key: to prefill in private key execution
        :param ~eth_account.account.Account account: the key-unaware management API
        """
        self._publicapi: AccountLocalActions = account

        self._address: ChecksumAddress = key.public_key.to_checksum_address()

        key_raw: bytes = key.to_bytes()
        self._private_key = key_raw

        self._key_obj: PrivateKey = key

    @property
    def address(self) -> ChecksumAddress:
        return self._address

    @property
    def key(self) -> bytes:
        """
        Get the private key.
        """
        return self._private_key

    def encrypt(
        self,
        password: str,
        kdf: Optional[KDFType] = None,
        iterations: Optional[int] = None,
    ) -> Dict[str, Any]:
        """
        Generate a string with the encrypted key.

        This uses the same structure as in
        :meth:`~eth_account.account.Account.encrypt`, but without a
        private key argument.
        """
        return self._publicapi.encrypt(
            self.key, password, kdf=kdf, iterations=iterations
        )

    def unsafe_sign_hash(self, message_hash: Hash32) -> SignedMessage:
        return cast(
            SignedMessage,
            self._publicapi.unsafe_sign_hash(
                message_hash,
                private_key=self.key,
            ),
        )

    def sign_message(self, signable_message: SignableMessage) -> SignedMessage:
        """
        Generate a string with the encrypted key.

        This uses the same structure as in
        :meth:`~eth_account.account.Account.sign_message`, but without a
        private key argument.
        """
        return cast(
            SignedMessage,
            self._publicapi.sign_message(signable_message, private_key=self.key),
        )

    def sign_transaction(
        self, transaction_dict: TransactionDictType, blobs: Optional[Blobs] = None
    ) -> SignedTransaction:
        return cast(
            SignedTransaction,
            self._publicapi.sign_transaction(transaction_dict, self.key, blobs=blobs),
        )

    def __bytes__(self) -> bytes:
        return self.key



from typing import (
    Optional,
    Tuple,
    cast,
)

from eth_keys.datatypes import (
    PrivateKey,
)
from eth_utils import (
    to_bytes,
    to_int,
)
from eth_utils.toolz import (
    pipe,
)

from eth_account._utils.legacy_transactions import (
    ChainAwareUnsignedTransaction,
    Transaction,
    UnsignedTransaction,
    encode_transaction,
    serializable_unsigned_transaction_from_dict,
    strip_signature,
)
from eth_account.typed_transactions import (
    TypedTransaction,
)
from eth_account.types import (
    Blobs,
    Bytes32,
    TransactionDictType,
)

CHAIN_ID_OFFSET = 35
V_OFFSET = 27

# signature versions
PERSONAL_SIGN_VERSION = b"E"  # Hex value 0x45
INTENDED_VALIDATOR_SIGN_VERSION = b"\x00"  # Hex value 0x00
STRUCTURED_DATA_SIGN_VERSION = b"\x01"  # Hex value 0x01


def sign_transaction_dict(
    eth_key: PrivateKey,
    transaction_dict: TransactionDictType,
    blobs: Optional[Blobs] = None,
) -> Tuple[int, int, int, bytes]:
    # generate RLP-serializable transaction, with defaults filled
    unsigned_transaction = serializable_unsigned_transaction_from_dict(
        transaction_dict, blobs=blobs
    )

    transaction_hash = unsigned_transaction.hash()

    # detect chain
    if isinstance(unsigned_transaction, UnsignedTransaction):
        chain_id = None
        (v, r, s) = sign_transaction_hash(eth_key, transaction_hash, chain_id)
    elif isinstance(unsigned_transaction, Transaction):
        chain_id = unsigned_transaction.v
        (v, r, s) = sign_transaction_hash(eth_key, transaction_hash, chain_id)
    elif isinstance(unsigned_transaction, TypedTransaction):
        # Each transaction type dictates its payload, and consequently,
        # all the funky logic around the `v` signature field is both obsolete &&
        # incorrect. We want to obtain the raw `v` and delegate
        # to the transaction type itself.
        (v, r, s) = eth_key.sign_msg_hash(transaction_hash).vrs
    else:
        # Cannot happen, but better for code to be defensive + self-documenting.
        raise TypeError(f"unknown Transaction object: {type(unsigned_transaction)}")

    # serialize transaction with rlp
    encoded_transaction = encode_transaction(unsigned_transaction, vrs=(v, r, s))

    return (v, r, s, encoded_transaction)


def hash_of_signed_transaction(txn_obj: Transaction) -> Bytes32:
    """
    Regenerate the hash of the signed transaction object.

    1. Infer the chain ID from the signature
    2. Strip out signature from transaction
    3. Annotate the transaction with that ID, if available
    4. Take the hash of the serialized, unsigned, chain-aware transaction

    Chain ID inference and annotation is according to EIP-155
    See details at https://github.com/ethereum/EIPs/blob/master/EIPS/eip-155.md

    :return: the hash of the provided transaction, to be signed
    """  # blocklint: URL pragma
    (chain_id, _v) = extract_chain_id(txn_obj.v)
    unsigned_parts = strip_signature(txn_obj)
    if chain_id is None:
        signable_transaction = UnsignedTransaction(*unsigned_parts)
    else:
        extended_transaction = unsigned_parts + [chain_id, 0, 0]
        signable_transaction = ChainAwareUnsignedTransaction(*extended_transaction)
    return signable_transaction.hash()


def extract_chain_id(raw_v: int) -> Tuple[Optional[int], int]:
    """
    Extracts chain ID, according to EIP-155.

    @return (chain_id, v)
    """
    above_id_offset = raw_v - CHAIN_ID_OFFSET
    if above_id_offset < 0:
        if raw_v in {0, 1}:
            return (None, raw_v + V_OFFSET)
        elif raw_v in {27, 28}:
            return (None, raw_v)
        else:
            raise ValueError(
                f"v {repr(raw_v)} is invalid, must be one of: 0, 1, 27, 28, 35+"
            )
    else:
        (chain_id, v_bit) = divmod(above_id_offset, 2)
        return (chain_id, v_bit + V_OFFSET)


def to_standard_signature_bytes(ethereum_signature_bytes: bytes) -> bytes:
    rs = ethereum_signature_bytes[:-1]
    v = to_int(ethereum_signature_bytes[-1])
    standard_v = to_standard_v(v)
    return rs + to_bytes(standard_v)


def to_standard_v(enhanced_v: int) -> int:
    (_chain, chain_naive_v) = extract_chain_id(enhanced_v)
    v_standard = chain_naive_v - V_OFFSET
    assert v_standard in {0, 1}
    return v_standard


def to_eth_v(v_raw: int, chain_id: Optional[int] = None) -> int:
    if chain_id is None:
        v = v_raw + V_OFFSET
    else:
        v = v_raw + CHAIN_ID_OFFSET + 2 * chain_id
    return v


def sign_transaction_hash(
    account: PrivateKey, transaction_hash: Bytes32, chain_id: Optional[int] = None
) -> Tuple[int, int, int]:
    signature = account.sign_msg_hash(transaction_hash)
    (v_raw, r, s) = signature.vrs
    v = to_eth_v(v_raw, chain_id)
    return (v, r, s)


def _pad_to_eth_word(bytes_val: bytes) -> Bytes32:
    return bytes_val.rjust(32, b"\0")


def to_bytes32(val: int) -> Bytes32:
    return cast(
        bytes,
        pipe(
            val,
            to_bytes,
            _pad_to_eth_word,
        ),
    )


def sign_message_hash(
    key: PrivateKey, msg_hash: Bytes32
) -> Tuple[int, int, int, bytes]:
    signature = key.sign_msg_hash(msg_hash)
    (v_raw, r, s) = signature.vrs
    v = to_eth_v(v_raw)
    eth_signature_bytes = to_bytes32(r) + to_bytes32(s) + to_bytes(v)
    return (v, r, s, eth_signature_bytes)



import hashlib
import hmac
from typing import (
    Union,
)
import unicodedata

from eth_keys import (
    keys,
)
from eth_utils import (
    ValidationError,
)
from hexbytes import (
    HexBytes,
)

PBKDF2_ROUNDS = 2048
SECP256K1_N = int(
    "FFFFFFFF_FFFFFFFF_FFFFFFFF_FFFFFFFE_BAAEDCE6_AF48A03B_BFD25E8C_D0364141", 16
)


def normalize_string(txt: Union[str, bytes]) -> str:
    if isinstance(txt, bytes):
        utxt = txt.decode("utf8")
    elif isinstance(txt, str):
        utxt = txt
    else:
        raise ValidationError("String value expected")

    return unicodedata.normalize("NFKD", utxt)


def sha256(data: bytes) -> bytes:
    return hashlib.sha256(data).digest()


def hmac_sha512(chain_code: bytes, data: bytes) -> bytes:
    """
    As specified by RFC4231 - https://tools.ietf.org/html/rfc4231 .
    """
    return hmac.new(chain_code, data, hashlib.sha512).digest()


def pbkdf2_hmac_sha512(passcode: str, salt: str) -> bytes:
    return hashlib.pbkdf2_hmac(
        "sha512",
        passcode.encode("utf-8"),
        salt.encode("utf-8"),
        PBKDF2_ROUNDS,
    )


def ec_point(pkey: bytes) -> bytes:
    """
    Compute `point(p)`, where `point` is ecdsa point multiplication.

    Note: Result is ecdsa public key serialized to compressed form
    """
    return keys.PrivateKey(HexBytes(pkey)).public_key.to_compressed_bytes()



from typing import (
    Any,
    NamedTuple,
    SupportsIndex,
    Tuple,
    Union,
    overload,
)

from hexbytes import (
    HexBytes,
)


class SignedTransaction(
    NamedTuple(
        "SignedTransaction",
        [
            ("raw_transaction", HexBytes),
            ("hash", HexBytes),
            ("r", int),
            ("s", int),
            ("v", int),
        ],
    )
):
    @overload
    def __getitem__(self, index: SupportsIndex) -> Any:
        ...

    @overload
    def __getitem__(self, index: slice) -> Tuple[Any, ...]:
        ...

    @overload
    def __getitem__(self, index: str) -> Any:
        ...

    def __getitem__(self, index: Union[SupportsIndex, slice, str]) -> Any:
        if isinstance(index, (int, slice)):
            return super().__getitem__(index)
        elif isinstance(index, str):
            return getattr(self, index)
        else:
            raise TypeError("Index must be an integer, slice, or string")


class SignedMessage(
    NamedTuple(
        "SignedMessage",
        [
            ("message_hash", HexBytes),
            ("r", int),
            ("s", int),
            ("v", int),
            ("signature", HexBytes),
        ],
    )
):
    @overload
    def __getitem__(self, index: SupportsIndex) -> Any:
        ...

    @overload
    def __getitem__(self, index: slice) -> Tuple[Any, ...]:
        ...

    @overload
    def __getitem__(self, index: str) -> Any:
        ...

    def __getitem__(self, index: Union[SupportsIndex, slice, str]) -> Any:
        if isinstance(index, (int, slice)):
            return super().__getitem__(index)
        elif isinstance(index, str):
            return getattr(self, index)
        else:
            raise TypeError("Index must be an integer, slice, or string")



import json
import pytest

from eth_utils import (
    ValidationError,
)

from eth_account.messages import (
    encode_typed_data,
)
from tests.eip712_messages import (
    ALL_VALID_EIP712_MESSAGES,
    INVALID,
    ONE_ARG_INVALID,
    convert_to_3_arg,
)


@pytest.mark.parametrize(
    "test_cases", (ALL_VALID_EIP712_MESSAGES, INVALID, ONE_ARG_INVALID)
)
def test_there_are_no_duplicate_test_cases(test_cases):
    string_test_cases = [json.dumps(msg, sort_keys=True) for msg in test_cases.values()]
    assert len(string_test_cases) == len(set(string_test_cases))


@pytest.mark.parametrize("message", ALL_VALID_EIP712_MESSAGES)
def test_valid_messages(message):
    assert encode_typed_data(
        full_message=ALL_VALID_EIP712_MESSAGES[message]
    ) == encode_typed_data(*convert_to_3_arg(ALL_VALID_EIP712_MESSAGES[message]))


@pytest.mark.parametrize("message", INVALID)
def test_invalid_messages(message):
    with pytest.raises(ValueError):
        encode_typed_data(full_message=INVALID[message])

    with pytest.raises(ValueError):
        encode_typed_data(*convert_to_3_arg(INVALID[message]))


@pytest.mark.parametrize("message", ONE_ARG_INVALID)
def test_messages_that_are_only_invalid_for_one_arg_encoding(message):
    with pytest.raises(ValidationError):
        encode_typed_data(full_message=ONE_ARG_INVALID[message])
    encode_typed_data(*convert_to_3_arg(ONE_ARG_INVALID[message]))



import pytest

from eth_utils import (
    ValidationError,
)

from eth_account.messages import (
    SignableMessage,
    encode_intended_validator,
)


@pytest.mark.parametrize(
    "primitive, hexstr, text, validator_address, expected_signable",
    (
        (b"", None, None, b"\xff" * 20, SignableMessage(b"\0", b"\xff" * 20, b"")),
        (
            b"",
            None,
            None,
            "0xFFfFfFffFFfffFFfFFfFFFFFffFFFffffFfFFFfF",
            SignableMessage(b"\0", b"\xff" * 20, b""),
        ),
        (None, "0x", None, b"\xff" * 20, SignableMessage(b"\0", b"\xff" * 20, b"")),
        (None, "", None, b"\xff" * 20, SignableMessage(b"\0", b"\xff" * 20, b"")),
        (None, None, "0x", b"\xff" * 20, SignableMessage(b"\0", b"\xff" * 20, b"0x")),
        (None, None, "", b"\xff" * 20, SignableMessage(b"\0", b"\xff" * 20, b"")),
    ),
)
def test_encode_intended_validator(
    primitive, hexstr, text, validator_address, expected_signable
):
    signable_message = encode_intended_validator(
        validator_address,
        primitive,
        hexstr=hexstr,
        text=text,
    )
    assert signable_message == expected_signable


@pytest.mark.parametrize(
    "invalid_address",
    (
        b"",
        "",
        None,
        # must be checksummed if hex-encoded
        "0xffffffffffffffffffffffffffffffffffffffff",
        b"\xff" * 19,
        b"\xff" * 21,
    ),
)
def test_encode_intended_validator_invalid_address(invalid_address):
    with pytest.raises(ValidationError):
        encode_intended_validator(invalid_address, b"")



from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    Any,
    Dict,
    Optional,
    Union,
)

from eth_keyfile.keyfile import (
    KDFType,
)
from eth_typing import (
    HexStr,
)
from eth_utils.curried import (
    combomethod,
)

from eth_account.datastructures import (
    SignedMessage,
    SignedTransaction,
)
from eth_account.messages import (
    SignableMessage,
)
from eth_account.types import (
    Blobs,
    PrivateKeyType,
    TransactionDictType,
)


class AccountLocalActions(ABC):
    @classmethod
    @abstractmethod
    def encrypt(
        self,
        private_key: PrivateKeyType,
        password: str,
        kdf: Optional[KDFType] = None,
        iterations: Optional[int] = None,
    ) -> Dict[str, Any]:
        pass

    @combomethod
    @abstractmethod
    def unsafe_sign_hash(
        self,
        message_hash: Union[HexStr, bytes, int],
        private_key: PrivateKeyType,
    ) -> SignedMessage:
        pass

    @combomethod
    @abstractmethod
    def sign_message(
        self,
        signable_message: SignableMessage,
        private_key: PrivateKeyType,
    ) -> SignedMessage:
        pass

    @combomethod
    @abstractmethod
    def sign_transaction(
        self,
        transaction_dict: TransactionDictType,
        private_key: PrivateKeyType,
        blobs: Optional[Blobs] = None,
    ) -> SignedTransaction:
        pass



import os
from typing import (
    Any,
    Optional,
    cast,
)

from eth_keyfile.keyfile import (
    KDFType,
)
from eth_utils import (
    is_binary_address,
    is_checksum_address,
    is_dict,
    is_hexstr,
)
from eth_utils.curried import (
    apply_one_of_formatters,
    hexstr_if_str,
    is_0x_prefixed,
    is_address,
    is_bytes,
    is_integer,
    is_list_like,
    is_string,
    to_bytes,
    to_int,
)
from eth_utils.toolz import (
    curry,
    identity,
)
from hexbytes import (
    HexBytes,
)

VALID_EMPTY_ADDRESSES = {None, b"", ""}


def is_none(val: Any) -> bool:
    return val is None


def is_valid_address(value: Any) -> bool:
    return is_binary_address(value) or is_checksum_address(value)


def is_int_or_prefixed_hexstr(val: Any) -> bool:
    if is_integer(val):
        return True
    elif isinstance(val, str) and is_0x_prefixed(val):
        return True
    else:
        return False


def is_empty_or_checksum_address(val: Any) -> bool:
    if val in VALID_EMPTY_ADDRESSES:
        return True
    else:
        return is_valid_address(val)


def is_rpc_structured_access_list(val: Any) -> bool:
    """Returns true if 'val' is a valid JSON-RPC structured access list."""
    if not is_list_like(val):
        return False
    for d in val:
        if not is_dict(d):
            return False
        if len(d) != 2:
            return False
        address = d.get("address")
        storage_keys = d.get("storageKeys")
        if any(_ is None for _ in (address, storage_keys)):
            return False
        if not is_address(address):
            return False
        if not is_list_like(storage_keys):
            return False
        for storage_key in storage_keys:
            if not is_int_or_prefixed_hexstr(storage_key):
                return False
    return True


def is_rlp_structured_access_list(val: Any) -> bool:
    """Returns true if 'val' is a valid rlp-structured access list."""
    if not is_list_like(val):
        return False
    for item in val:
        if not is_list_like(item):
            return False
        if len(item) != 2:
            return False
        address, storage_keys = item
        if not is_address(address):
            return False
        for storage_key in storage_keys:
            if not is_int_or_prefixed_hexstr(storage_key):
                return False
    return True


# type ignored because curry doesn't preserve typing
@curry  # type: ignore[misc]
def is_sequence_of_bytes_or_hexstr(
    value: Any, item_bytes_size: Optional[int] = None, can_be_empty: bool = False
) -> bool:
    if not is_list_like(value):
        return False

    if not can_be_empty and len(value) == 0:
        return False

    if not all(is_bytes(item) or is_hexstr(item) for item in value):
        return False

    if item_bytes_size is not None and not all(
        len(HexBytes(item)) == item_bytes_size for item in value
    ):
        return False

    return True


LEGACY_TRANSACTION_FORMATTERS = {
    "nonce": hexstr_if_str(to_int),
    "gasPrice": hexstr_if_str(to_int),
    "gas": hexstr_if_str(to_int),
    "to": apply_one_of_formatters(
        (
            (is_string, hexstr_if_str(to_bytes)),
            (is_bytes, identity),
            (is_none, lambda val: b""),
        )
    ),
    "value": hexstr_if_str(to_int),
    "data": hexstr_if_str(to_bytes),
    "v": hexstr_if_str(to_int),
    "r": hexstr_if_str(to_int),
    "s": hexstr_if_str(to_int),
}

LEGACY_TRANSACTION_VALID_VALUES = {
    "nonce": is_int_or_prefixed_hexstr,
    "gasPrice": is_int_or_prefixed_hexstr,
    "gas": is_int_or_prefixed_hexstr,
    "to": is_empty_or_checksum_address,
    "value": is_int_or_prefixed_hexstr,
    "data": lambda val: isinstance(val, (int, str, bytes, bytearray)),
    "chainId": lambda val: val is None or is_int_or_prefixed_hexstr(val),
}


def validate_and_set_default_kdf() -> KDFType:
    os_kdf = os.getenv("ETH_ACCOUNT_KDF", "scrypt")
    if os_kdf not in ("pbkdf2", "scrypt"):
        raise ValueError(
            f"Invalid KDF type: {os_kdf}. Must be one of 'pbkdf2' or 'scrypt'"
        )
    return cast(KDFType, os_kdf)



import os
import pytest

from eth_keyfile.keyfile import (
    get_default_work_factor_for_kdf,
)
from eth_keys import (
    keys,
)
from eth_utils import (
    is_checksum_address,
    to_bytes,
    to_hex,
    to_int,
)
from eth_utils.toolz import (
    dissoc,
)
from hexbytes import (
    HexBytes,
)
from hypothesis import (
    given,
    strategies as st,
)

from eth_account import (
    Account,
)
from eth_account.messages import (
    defunct_hash_message,
    encode_defunct,
    encode_intended_validator,
    encode_typed_data,
)
from tests.eip712_messages import (
    ALL_VALID_EIP712_MESSAGES,
)

# from https://github.com/ethereum/tests/blob/3930ca3a9a377107d5792b3e7202f79c688f1a67/BasicTests/txtest.json # noqa: 501
ETH_TEST_TRANSACTIONS = [
    {
        "chainId": None,
        "key": "c85ef7d79691fe79573b1a7064c19c1a9819ebdbd1faaab1a8ec92344438aaf4",
        "nonce": 0,
        "gasPrice": 1000000000000,
        "gas": 10000,
        "to": "0x13978aee95f38490e9769C39B2773Ed763d9cd5F",
        "value": 10000000000000000,
        "data": "",
        "unsigned": "eb8085e8d4a510008227109413978aee95f38490e9769c39b2773ed763d9cd5f872386f26fc1000080808080",  # noqa: 501
        "signed": "f86b8085e8d4a510008227109413978aee95f38490e9769c39b2773ed763d9cd5f872386f26fc10000801ba0eab47c1a49bf2fe5d40e01d313900e19ca485867d462fe06e139e3a536c6d4f4a014a569d327dcda4b29f74f93c0e9729d2f49ad726e703f9cd90dbb0fbf6649f1",  # noqa: 501
    },
    {
        "chainId": None,
        "key": "c87f65ff3f271bf5dc8643484f66b200109caffe4bf98c4cb393dc35740b28c0",
        "nonce": 0,
        "gasPrice": 1000000000000,
        "gas": 10000,
        "to": "",
        "value": 0,
        "data": "6025515b525b600a37f260003556601b596020356000355760015b525b54602052f260255860005b525b54602052f2",  # noqa: 501
        "unsigned": "f83f8085e8d4a510008227108080af6025515b525b600a37f260003556601b596020356000355760015b525b54602052f260255860005b525b54602052f2808080",  # noqa: 501
        "signed": "f87f8085e8d4a510008227108080af6025515b525b600a37f260003556601b596020356000355760015b525b54602052f260255860005b525b54602052f21ba05afed0244d0da90b67cf8979b0f246432a5112c0d31e8d5eedd2bc17b171c694a0bb1035c834677c2e1185b8dc90ca6d1fa585ab3d7ef23707e1a497a98e752d1b",  # noqa: 501
    },
    {
        "chainId": None,
        "key": "c85ef7d79691fe79573b1a7064c19c1a9819ebdbd1faaab1a8ec92344438aaf4",
        "nonce": 0,
        "gasPrice": 1000000000000,
        "gas": 10000,
        "to": HexBytes("0x13978aee95f38490e9769C39B2773Ed763d9cd5F"),
        "value": 10000000000000000,
        "data": "",
        "unsigned": "eb8085e8d4a510008227109413978aee95f38490e9769c39b2773ed763d9cd5f872386f26fc1000080808080",  # noqa: 501
        "signed": "f86b8085e8d4a510008227109413978aee95f38490e9769c39b2773ed763d9cd5f872386f26fc10000801ba0eab47c1a49bf2fe5d40e01d313900e19ca485867d462fe06e139e3a536c6d4f4a014a569d327dcda4b29f74f93c0e9729d2f49ad726e703f9cd90dbb0fbf6649f1",  # noqa: 501
    },
    # Typed Transaction (EIP-2930's access list transaction) - empty list.
    {
        "key": "fad9c8855b740a0b7ed4c221dbad0f33a83a49cad6b3fe8d5817ac83d38b6a19",
        "gas": "0x186a0",
        "gasPrice": "0x3b9aca00",
        "data": "0x616263646566",
        "nonce": "0x27",
        "to": "0x09616C3d61b3331fc4109a9E41a8BDB7d9776609",
        "value": "0x5af3107a4000",
        "type": "0x1",
        "accessList": [],
        "chainId": "0x76c",
        "signed": "0x01f87482076c27843b9aca00830186a09409616c3d61b3331fc4109a9e41a8bdb7d9776609865af3107a400086616263646566c080a0bad1a40fa2d90dc7539831bb82dfccf9b7094eab238d50c4369b805fb7241c58a046ab7eb7ff8cdfd203847b7e1b2f9e41208bba76a86ae3eeb97fe2727763aa12",  # noqa: 501
    },
    # Typed Transaction (EIP-2930's access list transaction) - non-empty list.
    {
        "key": "fad9c8855b740a0b7ed4c221dbad0f33a83a49cad6b3fe8d5817ac83d38b6a19",
        "gas": "0x186a0",
        "gasPrice": "0x3b9aca00",
        "data": "0x616263646566",
        "nonce": "0x22",
        "to": "0x09616C3d61b3331fc4109a9E41a8BDB7d9776609",
        "value": "0x5af3107a4000",
        "type": "0x1",
        "accessList": [
            {
                "address": "0x0000000000000000000000000000000000000001",
                "storageKeys": [
                    "0x0100000000000000000000000000000000000000000000000000000000000000"
                ],
            },
        ],
        "chainId": "0x76c",
        "signed": "0x01f8ad82076c22843b9aca00830186a09409616c3d61b3331fc4109a9e41a8bdb7d9776609865af3107a400086616263646566f838f7940000000000000000000000000000000000000001e1a0010000000000000000000000000000000000000000000000000000000000000001a08289e85fa00f8f7f78a53cf147a87b2a7f0d27e64d7571f9d06a802e365c3430a017dc77eae36c88937db4a5179f57edc6119701652f3f1c6f194d1210d638a061",  # noqa: 501
    },
    # Typed Transaction (EIP-1559's dynamic fee transaction)
    {
        "key": "70af7ec25374c2b06cbfffeaf9817a3d1bc61854abbf39e9e00de734b0c0c8c4",
        "gas": "0x186a0",
        "maxFeePerGas": "0x77359400",
        "maxPriorityFeePerGas": "0x77359400",
        "data": "0x5544",
        "nonce": "0x2",
        "to": "0x96216849c49358B10257cb55b28eA603c874b05E",
        "value": "0x5af3107a4000",
        "type": "0x2",
        "chainId": "0x539",
        "signed": "0x02f8758205390284773594008477359400830186a09496216849c49358b10257cb55b28ea603c874b05e865af3107a4000825544c001a008badd022b23e5bb51a13a18c99185dbd24aac054f876a3cf89a61672e998b89a01d14fb34a688d516f6444fdb6ceca7778be786cc70d3fa29714b6f232f09a83a",  # noqa: 501
    },
    # Typed Transaction (EIP-1559's dynamic fee transaction) -
    # integer values and access list
    {
        "key": "70af7ec25374c2b06cbfffeaf9817a3d1bc61854abbf39e9e00de734b0c0c8c4",
        "gas": 100000,
        "maxFeePerGas": 2000000000,
        "maxPriorityFeePerGas": 1000000000,
        "data": "0x5544",
        "nonce": "0x2",
        "to": "0x96216849c49358B10257cb55b28eA603c874b05E",
        "value": "0x5af3107a4000",
        "type": "0x2",
        "accessList": [
            {
                "address": "0x0000000000000000000000000000000000000001",
                "storageKeys": [
                    "0x0100000000000000000000000000000000000000000000000000000000000000"
                ],
            },
        ],
        "chainId": "0x539",
        "signed": "0x02f8ae82053902843b9aca008477359400830186a09496216849c49358b10257cb55b28ea603c874b05e865af3107a4000825544f838f7940000000000000000000000000000000000000001e1a0010000000000000000000000000000000000000000000000000000000000000001a036727d3bb8339377ac1d4c8b4d8c684a75d36047ebc369628be8d4d0f962d70ea064b1a8d2568bb9c44290242e8af47e1f428e2bf3073d9822a753649247321dfc",  # noqa: 501
    },
]


PRIVATE_KEY_AS_BYTES = b"unicorns" * 4
PRIVATE_KEY_AS_HEXSTR = (
    "0x756e69636f726e73756e69636f726e73756e69636f726e73756e69636f726e73"
)
PRIVATE_KEY_AS_INT = 0x756E69636F726E73756E69636F726E73756E69636F726E73756E69636F726E73
PRIVATE_KEY_AS_OBJ = keys.PrivateKey(PRIVATE_KEY_AS_BYTES)
ACCT_ADDRESS = "0xa79F6f349C853F9Ea0B29636779ae3Cb4E3BA729"

PRIVATE_KEY_AS_BYTES_ALT = b"rainbows" * 4
PRIVATE_KEY_AS_HEXSTR_ALT = (
    "0x7261696e626f77737261696e626f77737261696e626f77737261696e626f7773"
)
PRIVATE_KEY_AS_INT_ALT = (
    0x7261696E626F77737261696E626F77737261696E626F77737261696E626F7773
)
PRIVATE_KEY_AS_OBJ_ALT = keys.PrivateKey(PRIVATE_KEY_AS_BYTES_ALT)
ACCT_ADDRESS_ALT = "0xafd7f0E16A1814B854b45f551AFD493BE5F039F9"


@pytest.fixture(
    params=[
        PRIVATE_KEY_AS_INT,
        PRIVATE_KEY_AS_HEXSTR,
        PRIVATE_KEY_AS_BYTES,
        PRIVATE_KEY_AS_OBJ,
    ]
)  # noqa: 501
def PRIVATE_KEY(request):
    return request.param


@pytest.fixture(
    params=[
        PRIVATE_KEY_AS_INT_ALT,
        PRIVATE_KEY_AS_HEXSTR_ALT,
        PRIVATE_KEY_AS_BYTES_ALT,
        PRIVATE_KEY_AS_OBJ_ALT,
    ]
)  # noqa: 501
def PRIVATE_KEY_ALT(request):
    return request.param


@pytest.fixture(params=["instance", "class"])
def acct(request):
    if request.param == "instance":
        return Account()
    elif request.param == "class":
        return Account
    else:
        raise Exception(f"account invocation {request.param} is not supported")


@pytest.fixture(scope="module")
def keyed_acct():
    return Account.from_key(PRIVATE_KEY_AS_BYTES)


@pytest.fixture(params=("text", "primitive", "hexstr"))
def message_encodings(request):
    if request == "text":
        return {"text": "hello world"}
    elif request == "primitive":
        return {"primitive": b"hello world"}
    else:
        return {"hexstr": "68656c6c6f20776f726c64"}


def test_eth_account_default_kdf(acct, monkeypatch):
    assert os.getenv("ETH_ACCOUNT_KDF") is None
    assert acct._default_kdf == "scrypt"

    monkeypatch.setenv("ETH_ACCOUNT_KDF", "pbkdf2")
    assert os.getenv("ETH_ACCOUNT_KDF") == "pbkdf2"

    import importlib

    from eth_account import (
        account,
    )

    importlib.reload(account)
    assert account.Account._default_kdf == "pbkdf2"


def test_eth_account_create_variation(acct):
    account1 = acct.create()
    account2 = acct.create()
    assert account1 != account2


def test_eth_account_equality(acct, PRIVATE_KEY):
    acct1 = acct.from_key(PRIVATE_KEY)
    acct2 = acct.from_key(PRIVATE_KEY)
    assert acct1 == acct2


def test_eth_account_from_key_reproducible(acct, PRIVATE_KEY):
    account1 = acct.from_key(PRIVATE_KEY)
    account2 = acct.from_key(PRIVATE_KEY)
    assert bytes(account1) == PRIVATE_KEY_AS_BYTES
    assert bytes(account1) == bytes(account2)
    assert isinstance(str(account1), str)


def test_eth_account_from_key_diverge(acct, PRIVATE_KEY, PRIVATE_KEY_ALT):
    account1 = acct.from_key(PRIVATE_KEY)
    account2 = acct.from_key(PRIVATE_KEY_ALT)
    assert bytes(account2) == PRIVATE_KEY_AS_BYTES_ALT
    assert bytes(account1) != bytes(account2)


def test_eth_account_from_key_seed_restrictions(acct):
    with pytest.raises(ValueError):
        acct.from_key(b"")
    with pytest.raises(ValueError):
        acct.from_key(b"\xff" * 31)
    with pytest.raises(ValueError):
        acct.from_key(b"\xff" * 33)


def test_eth_account_from_key_properties(acct, PRIVATE_KEY):
    account = acct.from_key(PRIVATE_KEY)
    assert callable(account.sign_transaction)
    assert callable(account.sign_message)
    assert is_checksum_address(account.address)
    assert account.address == ACCT_ADDRESS
    assert account.key == PRIVATE_KEY_AS_OBJ


def test_eth_account_create_properties(acct):
    account = acct.create()
    assert callable(account.sign_transaction)
    assert callable(account.sign_message)
    assert is_checksum_address(account.address)
    assert isinstance(account.key, bytes) and len(account.key) == 32


def test_eth_account_recover_transaction_example(acct):
    raw_tx_hex = "0xf8640d843b9aca00830e57e0945b2063246f2191f18f2675cedb8b28102e957458018025a00c753084e5a8290219324c1a3a86d4064ded2d15979b1ea790734aaa2ceaafc1a0229ca4538106819fd3a5509dd383e8fe4b731c6870339556a5c06feb9cf330bb"  # noqa: E501
    from_account = acct.recover_transaction(raw_tx_hex)
    assert from_account == "0xFeC2079e80465cc8C687fFF9EE6386ca447aFec4"


def test_eth_account_recover_transaction_with_literal(acct):
    raw_tx = 0xF8640D843B9ACA00830E57E0945B2063246F2191F18F2675CEDB8B28102E957458018025A00C753084E5A8290219324C1A3A86D4064DED2D15979B1EA790734AAA2CEAAFC1A0229CA4538106819FD3A5509DD383E8FE4B731C6870339556A5C06FEB9CF330BB  # noqa: E501
    from_account = acct.recover_transaction(raw_tx)
    assert from_account == "0xFeC2079e80465cc8C687fFF9EE6386ca447aFec4"


def test_eth_account_recover_message(acct):
    v, r, s = (
        28,
        "0xe6ca9bba58c88611fad66a6ce8f996908195593807c4b38bd528d2cff09d4eb3",
        "0x3e5bfbbf4d3e39b1a2fd816a7680c19ebebaf3a141b239934ad43cb33fcec8ce",
    )
    message_text = "I♥SF"
    message = encode_defunct(text=message_text)
    from_account = acct.recover_message(message, vrs=(v, r, s))
    assert from_account == "0x5ce9454909639D2D17A3F753ce7d93fa0b9aB12E"


@pytest.mark.parametrize(
    "signature_bytes",
    [
        # test signature bytes with standard v (0 in this case)
        b"\0Q[\xc8\xfd2&N!\xec\x08 \xe8\xc5\x12>\xd5\x8c\x11\x95\xc9\xea\x17\xcb\x01\x8b\x1a\xd4\x07<\xc5\xa6\0\x80\xf5\xdc\xec9zZ\x8cR0\x82\xbf\xa4\x17qV\x89\x03\xaaUN\xc0k\xa8G\\\xa9\x05\x0f\xb7\xd5\x00",  # noqa: E501
        # test signature bytes with chain-naive v (27 in this case)
        b"\0Q[\xc8\xfd2&N!\xec\x08 \xe8\xc5\x12>\xd5\x8c\x11\x95\xc9\xea\x17\xcb\x01\x8b\x1a\xd4\x07<\xc5\xa6\0\x80\xf5\xdc\xec9zZ\x8cR0\x82\xbf\xa4\x17qV\x89\x03\xaaUN\xc0k\xa8G\\\xa9\x05\x0f\xb7\xd5\x1b",  # noqa: E501
    ],
    ids=["test_sig_bytes_standard_v", "test_sig_bytes_chain_naive_v"],
)
def test_eth_account_recover_signature_bytes(acct, signature_bytes):
    # found a signature with a leading 0 byte in both r and s
    message = encode_defunct(text="10284")
    from_account = acct.recover_message(message, signature=signature_bytes)
    assert from_account == "0x2c7536E3605D9C16a7a3D7b1898e529396a65c23"


@pytest.mark.parametrize("raw_v", (0, 27))
@pytest.mark.parametrize("as_hex", (False, True))
def test_eth_account_recover_vrs(acct, raw_v, as_hex):
    # found a signature with a leading 0 byte in both r and s
    raw_r, raw_s = (
        143748089818580655331728101695676826715814583506606354117109114714663470502,
        227853308212209543997879651656855994238138056366857653269155208245074180053,
    )
    if as_hex:
        vrs = map(to_hex, (raw_v, raw_r, raw_s))
    else:
        vrs = raw_v, raw_r, raw_s

    message = encode_defunct(text="10284")
    from_account = acct.recover_message(message, vrs=vrs)
    assert from_account == "0x2c7536E3605D9C16a7a3D7b1898e529396a65c23"


@pytest.mark.parametrize(
    "message, expected",
    [
        (
            "Message tö sign. Longer than hash!",
            HexBytes(
                "0x10c7cb57942998ab214c062e7a57220a174aacd80418cead9f90ec410eacada1"
            ),
        ),
        (
            # Intentionally sneaky: message is a hexstr interpreted as text
            "0x4d6573736167652074c3b6207369676e2e204c6f6e676572207468616e206861736821",
            HexBytes(
                "0x6192785e9ad00100e7332ff585824b65eafa30bc8f1265cf86b5368aa3ab5d56"
            ),
        ),
        (
            "Hello World",
            HexBytes(
                "0xa1de988600a42c4b4ab089b619297c17d53cffae5d5120d82d8a92d0bb3b78f2"
            ),
        ),
    ],
    ids=["message_to_sign", "hexstr_as_text", "hello_world"],
)
def test_eth_account_hash_message_text(message, expected):
    assert defunct_hash_message(text=message) == expected


@pytest.mark.parametrize(
    "message, expected",
    [
        (
            "0x4d6573736167652074c3b6207369676e2e204c6f6e676572207468616e206861736821",
            HexBytes(
                "0x10c7cb57942998ab214c062e7a57220a174aacd80418cead9f90ec410eacada1"
            ),
        ),
        (
            "0x29d9f7d6a1d1e62152f314f04e6bd4300ad56fd72102b6b83702869a089f470c",
            HexBytes(
                "0xe709159ef0e6323c705786fc50e47a8143812e9f82f429e585034777c7bf530b"
            ),
        ),
    ],
    ids=["hexbytes_1", "hexbytes_2"],
)
def test_eth_account_hash_message_hexstr(acct, message, expected):
    assert defunct_hash_message(hexstr=message) == expected


@given(st.text())
def test_sign_message_against_sign_hash_as_text(keyed_acct, message_text):
    # sign via hash
    msg_hash = defunct_hash_message(text=message_text)
    signed_via_hash = keyed_acct.unsafe_sign_hash(msg_hash)

    # sign via message
    signable_message = encode_defunct(text=message_text)
    signed_via_message = keyed_acct.sign_message(signable_message)
    assert signed_via_hash == signed_via_message


@given(st.binary())
def test_sign_message_against_sign_hash_as_bytes(keyed_acct, message_bytes):
    # sign via hash
    msg_hash = defunct_hash_message(message_bytes)
    signed_via_hash = keyed_acct.unsafe_sign_hash(msg_hash)

    # sign via message
    signable_message = encode_defunct(message_bytes)
    signed_via_message = keyed_acct.sign_message(signable_message)

    assert signed_via_hash == signed_via_message


@given(st.binary())
def test_sign_message_against_sign_hash_as_hex(keyed_acct, message_bytes):
    message_hex = to_hex(message_bytes)

    # sign via hash
    msg_hash_hex = defunct_hash_message(hexstr=message_hex)
    signed_via_hash_hex = keyed_acct.unsafe_sign_hash(msg_hash_hex)

    # sign via message
    signable_message_hex = encode_defunct(hexstr=message_hex)
    signed_via_message_hex = keyed_acct.sign_message(signable_message_hex)

    assert signed_via_hash_hex == signed_via_message_hex


@pytest.mark.parametrize(
    "message, key, expected_bytes, expected_hash, v, r, s, signature",
    (
        (
            "Some data",
            "0x4c0883a69102937d6231471b5dbb6204fe5129617082792ae468d01a3f362318",
            b"Some data",
            HexBytes(
                "0x1da44b586eb0729ff70a73c326926f6ed5a25f5b056e7f47fbc6e58d86871655"
            ),
            28,
            83713930994764734002432606962255364472443135907807238282514898577139886061053,  # noqa: E501
            43435997768575461196683613590576722655951133545204789519877940758262837256233,  # noqa: E501
            HexBytes(
                "0xb91467e570a6466aa9e9876cbcd013baba02900b8979d43fe208a4a4f339f5fd6007e74cd82e037b800186422fc2da167c747ef045e5d18a5f5d4300f8e1a0291c"  # noqa: E501
            ),
        ),
        (
            "Some data",
            keys.PrivateKey(
                HexBytes(
                    "0x4c0883a69102937d6231471b5dbb6204fe5129617082792ae468d01a3f362318"
                )
            ),
            b"Some data",
            HexBytes(
                "0x1da44b586eb0729ff70a73c326926f6ed5a25f5b056e7f47fbc6e58d86871655"
            ),
            28,
            83713930994764734002432606962255364472443135907807238282514898577139886061053,  # noqa: E501
            43435997768575461196683613590576722655951133545204789519877940758262837256233,  # noqa: E501
            HexBytes(
                "0xb91467e570a6466aa9e9876cbcd013baba02900b8979d43fe208a4a4f339f5fd6007e74cd82e037b800186422fc2da167c747ef045e5d18a5f5d4300f8e1a0291c"  # noqa: E501
            ),
        ),
        (
            "10284",
            "0x4c0883a69102937d6231471b5dbb6204fe5129617082792ae468d01a3f362318",
            b"10284",
            HexBytes(
                "0x0a162a5efbba02f38db3114531c8acba39fe676f09f7e471d93e8a06c471821c"
            ),
            27,
            143748089818580655331728101695676826715814583506606354117109114714663470502,
            227853308212209543997879651656855994238138056366857653269155208245074180053,
            HexBytes(
                "0x00515bc8fd32264e21ec0820e8c5123ed58c1195c9ea17cb018b1ad4073cc5a60080f5dcec397a5a8c523082bfa41771568903aa554ec06ba8475ca9050fb7d51b"  # noqa: E501
            ),
        ),
    ),
    ids=[
        "web3js_hex_str_example",
        "web3js_eth_keys.datatypes.PrivateKey_example",
        "31byte_r_and_s",
    ],
)
def test_eth_account_sign(
    acct, message, key, expected_bytes, expected_hash, v, r, s, signature
):
    signable = encode_defunct(text=message)
    signed = acct.sign_message(signable, private_key=key)
    assert signed.message_hash == signed["message_hash"] == expected_hash
    assert signed.v == signed["v"] == v
    assert signed.r == signed["r"] == r
    assert signed.s == signed["s"] == s
    assert signed.signature == signed["signature"] == signature

    account = acct.from_key(key)
    assert account.sign_message(signable) == signed


def test_eth_valid_account_address_sign_data_with_intended_validator(
    acct, message_encodings
):
    account = acct.create()
    signable = encode_intended_validator(
        account.address,
        **message_encodings,
    )
    signed = account.sign_message(signable)
    signed_classmethod = acct.sign_message(signable, account.key)
    assert signed == signed_classmethod
    new_addr = acct.recover_message(signable, signature=signed.signature)
    assert new_addr == account.address


def test_eth_short_account_address_sign_data_with_intended_validator(
    acct, message_encodings
):
    account = acct.create()

    address_in_bytes = to_bytes(hexstr=account.address)
    # Test for all lengths of addresses < 20 bytes
    for i in range(1, 21):
        with pytest.raises(TypeError):
            # Raise TypeError if the address is less than 20 bytes
            defunct_hash_message(
                **message_encodings,
                signature_version=b"\x00",
                version_specific_data=to_hex(address_in_bytes[:-i]),
            )


def test_eth_long_account_address_sign_data_with_intended_validator(
    acct, message_encodings
):
    account = acct.create()

    address_in_bytes = to_bytes(hexstr=account.address)
    with pytest.raises(TypeError):
        # Raise TypeError if the address is more than 20 bytes
        defunct_hash_message(
            **message_encodings,
            signature_version=b"\x00",
            version_specific_data=to_hex(address_in_bytes + b"\x00"),
        )


@pytest.mark.parametrize(
    "txn, private_key, expected_raw_tx, tx_hash, r, s, v",
    (
        (
            {
                "to": "0xF0109fC8DF283027b6285cc889F5aA624EaC1F55",
                "value": 1000000000,
                "gas": 2000000,
                "gasPrice": 234567897654321,
                "nonce": 0,
                "chainId": 1,
            },
            "0x4c0883a69102937d6231471b5dbb6204fe5129617082792ae468d01a3f362318",
            HexBytes(
                "0xf86a8086d55698372431831e848094f0109fc8df283027b6285cc889f5aa624eac1f55843b9aca008025a009ebb6ca057a0535d6186462bc0b465b561c94a295bdb0621fc19208ab149a9ca0440ffd775ce91a833ab410777204d5341a6f9fa91216a6f3ee2c051fea6a0428"  # noqa: E501
            ),
            HexBytes(
                "0xd8f64a42b57be0d565f385378db2f6bf324ce14a594afc05de90436e9ce01f60"
            ),
            4487286261793418179817841024889747115779324305375823110249149479905075174044,  # noqa: E501
            30785525769477805655994251009256770582792548537338581640010273753578382951464,  # noqa: E501
            37,
        ),
        (
            {
                "to": "0xF0109fC8DF283027b6285cc889F5aA624EaC1F55",
                "value": 1000000000,
                "gas": 2000000,
                "gasPrice": 234567897654321,
                "nonce": 0,
                "chainId": 1,
            },
            keys.PrivateKey(
                HexBytes(
                    "0x4c0883a69102937d6231471b5dbb6204fe5129617082792ae468d01a3f362318"
                )
            ),  # noqa: E501
            HexBytes(
                "0xf86a8086d55698372431831e848094f0109fc8df283027b6285cc889f5aa624eac1f55843b9aca008025a009ebb6ca057a0535d6186462bc0b465b561c94a295bdb0621fc19208ab149a9ca0440ffd775ce91a833ab410777204d5341a6f9fa91216a6f3ee2c051fea6a0428"  # noqa: E501
            ),
            HexBytes(
                "0xd8f64a42b57be0d565f385378db2f6bf324ce14a594afc05de90436e9ce01f60"
            ),
            4487286261793418179817841024889747115779324305375823110249149479905075174044,  # noqa: E501
            30785525769477805655994251009256770582792548537338581640010273753578382951464,  # noqa: E501
            37,
        ),
        (
            {
                "to": "0xF0109fC8DF283027b6285cc889F5aA624EaC1F55",
                "value": 0,
                "gas": 31853,
                "gasPrice": 0,
                "nonce": 0,
                "chainId": 1,
            },
            "0x4c0883a69102937d6231471b5dbb6204fe5129617082792ae468d01a3f362318",
            HexBytes(
                "0xf85d8080827c6d94f0109fc8df283027b6285cc889f5aa624eac1f558080269f22f17b38af35286ffbb0c6376c86ec91c20ecbad93f84913a0cc15e7580cd99f83d6e12e82e3544cb4439964d5087da78f74cefeec9a450b16ae179fd8fe20"  # noqa: E501
            ),
            HexBytes(
                "0xb0c5e2c6b29eeb0b9c1d63eaa8b0f93c02ead18ae01cb7fc795b0612d3e9d55a"
            ),
            61739443115046231975538240097110168545680205678104352478922255527799426265,
            232940010090391255679819602567388136081614408698362277324138554019997613600,
            38,
        ),
        (
            {
                "gas": "0x186a0",
                "gasPrice": "0x3b9aca00",
                "data": "0x616263646566",
                "nonce": "0x22",
                "to": "0x09616C3d61b3331fc4109a9E41a8BDB7d9776609",
                "value": "0x5af3107a4000",
                "type": "0x1",
                "accessList": [
                    {
                        "address": "0x0000000000000000000000000000000000000001",
                        "storageKeys": [
                            "0x0100000000000000000000000000000000000000000000000000000000000000"  # noqa: E501
                        ],
                    },
                ],
                "chainId": "0x76c",
            },
            "0xfad9c8855b740a0b7ed4c221dbad0f33a83a49cad6b3fe8d5817ac83d38b6a19",
            HexBytes(
                "0x01f8ad82076c22843b9aca00830186a09409616c3d61b3331fc4109a9e41a8bdb7d9776609865af3107a400086616263646566f838f7940000000000000000000000000000000000000001e1a0010000000000000000000000000000000000000000000000000000000000000001a08289e85fa00f8f7f78a53cf147a87b2a7f0d27e64d7571f9d06a802e365c3430a017dc77eae36c88937db4a5179f57edc6119701652f3f1c6f194d1210d638a061"  # noqa: E501
            ),
            HexBytes(
                "0x2a791d5483705e444fa6d493e6f504836cf54ca78335c60dc81bcf320e95e49c"
            ),
            59044332146903025833144089863119240337233261477961028574753111682592582415408,  # noqa: E501
            10792729512059697976635619515571917958852106732672247829612911298843986403425,  # noqa: E501
            1,
        ),
    ),
    ids=[
        "web3js_hex_str_example",
        "web3js_eth_keys.datatypes.PrivateKey_example",
        "31byte_r_and_s",
        "access_list_tx",
    ],
)
def test_eth_account_sign_transaction(
    acct, txn, private_key, expected_raw_tx, tx_hash, r, s, v
):
    signed = acct.sign_transaction(txn, private_key)
    assert signed.r == signed["r"] == r
    assert signed.s == signed["s"] == s
    assert signed.v == signed["v"] == v
    assert signed.raw_transaction == signed["raw_transaction"] == expected_raw_tx
    assert signed.hash == signed["hash"] == tx_hash

    account = acct.from_key(private_key)
    assert account.sign_transaction(txn) == signed


@pytest.mark.parametrize(
    "transaction",
    ETH_TEST_TRANSACTIONS,
)
def test_eth_account_sign_transaction_from_eth_test(acct, transaction):
    expected_raw_txn = transaction["signed"]
    key = transaction["key"]

    unsigned_txn = dissoc(transaction, "key", "signed", "unsigned")

    # validate r, in order to validate the transaction hash
    # There is some ambiguity about whether `r` will always be deterministically
    # generated from the transaction hash and private key, mostly due to code
    # author's ignorance. The example test fixtures and implementations seem to
    # agree, so far. See ecdsa_raw_sign() in /eth_keys/backends/native/ecdsa.py
    signed = acct.sign_transaction(unsigned_txn, key)
    assert signed.r == to_int(hexstr=expected_raw_txn[-130:-66])

    # confirm that signed transaction can be recovered to the sender
    expected_sender = acct.from_key(key).address
    assert acct.recover_transaction(signed.raw_transaction) == expected_sender


@pytest.mark.parametrize(
    "transaction",
    ETH_TEST_TRANSACTIONS,
)
def test_eth_account_recover_transaction_from_eth_test(acct, transaction):
    raw_txn = transaction["signed"]
    expected_sender = acct.from_key(transaction["key"]).address
    assert acct.recover_transaction(raw_txn) == expected_sender


def get_encrypt_test_params():
    """
    Params for testing Account#encrypt. Due to not being able to provide fixtures to
    pytest.mark.parametrize, we opt for creating the params in a non-fixture method
    here instead of providing fixtures for the private key and password.
    """
    key = "0x4c0883a69102937d6231471b5dbb6204fe5129617082792ae468d01a3f362318"
    key_bytes = to_bytes(hexstr=key)
    private_key = keys.PrivateKey(HexBytes(key))
    password = "test!"

    # 'private_key, password, kdf, iterations, expected_decrypted_key, expected_kdf'
    return [
        (key, password, None, None, key_bytes, "scrypt"),
        (private_key, password, None, 2, private_key.to_bytes(), "scrypt"),
        (key, password, "pbkdf2", 4, key_bytes, "pbkdf2"),
        (key, password, None, 8, key_bytes, "scrypt"),
        (key, password, "pbkdf2", 16, key_bytes, "pbkdf2"),
        (key, password, "scrypt", 32, key_bytes, "scrypt"),
    ]


@pytest.mark.parametrize(
    "private_key, password, kdf, iterations, expected_decrypted_key, expected_kdf",
    get_encrypt_test_params(),
    ids=[
        "hex_str",
        "eth_keys.datatypes.PrivateKey",
        "hex_str_provided_kdf",
        "hex_str_default_kdf_provided_iterations",
        "hex_str_pbkdf2_provided_iterations",
        "hex_str_scrypt_provided_iterations",
    ],
)
def test_eth_account_encrypt(
    acct, private_key, password, kdf, iterations, expected_decrypted_key, expected_kdf
):
    if kdf is None:
        encrypted = acct.encrypt(private_key, password, iterations=iterations)
    else:
        encrypted = acct.encrypt(private_key, password, kdf=kdf, iterations=iterations)

    assert encrypted["address"] == "2c7536E3605D9C16a7a3D7b1898e529396a65c23"
    assert encrypted["version"] == 3
    assert encrypted["crypto"]["kdf"] == expected_kdf

    if iterations is None:
        expected_iterations = get_default_work_factor_for_kdf(expected_kdf)
    else:
        expected_iterations = iterations

    if expected_kdf == "pbkdf2":
        assert encrypted["crypto"]["kdfparams"]["c"] == expected_iterations
    elif expected_kdf == "scrypt":
        assert encrypted["crypto"]["kdfparams"]["n"] == expected_iterations
    else:
        raise Exception(
            f"test must be upgraded to confirm iterations with kdf {expected_kdf}"
        )

    decrypted_key = acct.decrypt(encrypted, password)

    assert decrypted_key == expected_decrypted_key


@pytest.mark.parametrize(
    "private_key, password, kdf, iterations, expected_decrypted_key, expected_kdf",
    get_encrypt_test_params(),
    ids=[
        "hex_str",
        "eth_keys.datatypes.PrivateKey",
        "hex_str_provided_kdf",
        "hex_str_default_kdf_provided_iterations",
        "hex_str_pbkdf2_provided_iterations",
        "hex_str_scrypt_provided_iterations",
    ],
)
def test_eth_account_prepared_encrypt(
    acct, private_key, password, kdf, iterations, expected_decrypted_key, expected_kdf
):
    account = acct.from_key(private_key)

    if kdf is None:
        encrypted = account.encrypt(password, iterations=iterations)
    else:
        encrypted = account.encrypt(password, kdf=kdf, iterations=iterations)

    assert encrypted["address"] == "2c7536E3605D9C16a7a3D7b1898e529396a65c23"
    assert encrypted["version"] == 3
    assert encrypted["crypto"]["kdf"] == expected_kdf

    if iterations is None:
        expected_iterations = get_default_work_factor_for_kdf(expected_kdf)
    else:
        expected_iterations = iterations

    if expected_kdf == "pbkdf2":
        assert encrypted["crypto"]["kdfparams"]["c"] == expected_iterations
    elif expected_kdf == "scrypt":
        assert encrypted["crypto"]["kdfparams"]["n"] == expected_iterations
    else:
        raise Exception(
            f"test must be upgraded to confirm iterations with kdf {expected_kdf}"
        )

    decrypted_key = acct.decrypt(encrypted, password)

    assert isinstance(decrypted_key, HexBytes)
    assert decrypted_key == expected_decrypted_key


@pytest.mark.parametrize("test_case", ALL_VALID_EIP712_MESSAGES)
def test_sign_typed_data_produces_same_result_as_encode_plus_sign(acct, test_case):
    test_message = ALL_VALID_EIP712_MESSAGES[test_case]
    encoded = encode_typed_data(full_message=test_message)
    signed_encoded = acct.sign_message(encoded, PRIVATE_KEY_AS_HEXSTR)
    assert (
        acct.sign_typed_data(PRIVATE_KEY_AS_HEXSTR, full_message=test_message)
        == signed_encoded
    )


@pytest.mark.parametrize(
    "domain_data, message_types, message_data, expected_sig",
    (
        (
            {
                "name": "Ether Mail",
                "version": "1",
                "chainId": 1,
                "verifyingContract": "0xCcCCccccCCCCcCCCCCCcCcCccCcCCCcCcccccccC",
            },
            {
                "Person": [
                    {"name": "name", "type": "string"},
                    {"name": "wallet", "type": "address"},
                ],
                "Mail": [
                    {"name": "from", "type": "Person"},
                    {"name": "to", "type": "Person"},
                    {"name": "contents", "type": "string"},
                ],
            },
            {
                "from": {
                    "name": "Cow",
                    "wallet": "0xCD2a3d9F938E13CD947Ec05AbC7FE734Df8DD826",
                },
                "to": {
                    "name": "Bob",
                    "wallet": "0xbBbBBBBbbBBBbbbBbbBbbbbBBbBbbbbBbBbbBBbB",
                },
                "contents": "Hello, Bob!",
            },
            HexBytes(
                "0x33600224dbc5a598b7c443b0cc9241b4b12ea10441244cc058442b31065b37232258c5d0b55f22362b1fa97d5c7fca4f40bfc1545417d974a22dfd4a8ac2de8b1b"  # noqa: E501
            ),
        ),
        (
            {
                "name": "Ether Mail",
                "version": "1",
                "chainId": 1,
                "verifyingContract": "0xCcCCccccCCCCcCCCCCCcCcCccCcCCCcCcccccccC",
            },
            {
                "Person": [
                    {"name": "name", "type": "string"},
                    {"name": "wallet", "type": "address"},
                ],
                "Mail": [
                    {"name": "from", "type": "Person"},
                    {"name": "to", "type": "Person"},
                    {"name": "cc", "type": "Person[]"},
                    {"name": "contents", "type": "string"},
                ],
            },
            {
                "from": {
                    "name": "Cow",
                    "wallet": "0xCD2a3d9F938E13CD947Ec05AbC7FE734Df8DD826",
                },
                "to": {
                    "name": "Bob",
                    "wallet": "0xbBbBBBBbbBBBbbbBbbBbbbbBBbBbbbbBbBbbBBbB",
                },
                "cc": [
                    {
                        "name": "Alice",
                        "wallet": "0xaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
                    },
                    {
                        "name": "Dot",
                        "wallet": "0xdddddddddddddddddddddddddddddddddddddddd",
                    },
                ],
                "contents": "Hello, Bob!",
            },
            HexBytes(
                "0xfcd28bfd425a26215d645481af446895f761b16896ac007c86433136ad9bb5c815ccdef9a6947396be54e92955c5c707ce65099d0371b4488143264675f5e18b1c"  # noqa: E501
            ),
        ),
    ),
)
def test_sign_typed_data_produces_expected_hash(
    acct, domain_data, message_types, message_data, expected_sig
):
    assert (
        acct.sign_typed_data(
            PRIVATE_KEY_AS_HEXSTR, domain_data, message_types, message_data
        ).signature
        == expected_sig
    )



"""
Generate Heirarchical Deterministic Wallets (HDWallet).

Partially implements the BIP-0032, BIP-0043, and BIP-0044 specifications:
BIP-0032: https://github.com/bitcoin/bips/blob/master/bip-0032.mediawiki
BIP-0043: https://github.com/bitcoin/bips/blob/master/bip-0043.mediawiki
BIP-0044: https://github.com/bitcoin/bips/blob/master/bip-0044.mediawiki

Skips serialization and public key derivation as unnecssary for this library's purposes.

Notes
-----

* Integers are modulo the order of the curve (referred to as n).
* Addition (+) of two coordinate pair is defined as application of
  the EC group operation.
* Concatenation (||) is the operation of appending one byte sequence onto another.


Definitions
-----------

* point(p): returns the coordinate pair resulting from EC point multiplication
  (repeated application of the EC group operation) of the secp256k1 base point
  with the integer p.
* ser_32(i): serialize a 32-bit unsigned integer i as a 4-byte sequence,
  most significant byte first.
* ser_256(p): serializes the integer p as a 32-byte sequence, most significant
  byte first.
* ser_P(P): serializes the coordinate pair P = (x,y) as a byte sequence using SEC1's
  compressed form: (0x02 or 0x03) || ser_256(x), where the header byte depends on the
  parity of the omitted y coordinate.
* parse_256(p): interprets a 32-byte sequence as a 256-bit number, most significant
  byte first.

"""  # blocklint: URL pragma
# Additional notes:
# - This module currently only implements private parent key => private child key
#   CKD function, as it is not necessary to the HD key derivation functions used
#   in this library to implement the other functions yet (as this module is only
#   used for derivation of private keys). That could change, but wasn't deemed
#   necessary at the time this module was introduced.
# - Unlike other libraries, this library does not use Bitcoin key serialization,
#   because it is not intended to be ultimately used for Bitcoin key derivations.
#   This presents a simplified API, and no expectation is given for `xpub/xpriv`
#   key derivation.
from typing import (
    Tuple,
    Type,
    Union,
)

from eth_utils import (
    ValidationError,
    to_int,
)

from ._utils import (
    SECP256K1_N,
    ec_point,
    hmac_sha512,
)

BASE_NODE_IDENTIFIERS = {"m", "M"}
HARD_NODE_SUFFIXES = {"'", "H"}


class Node(int):
    """
    A base node class.
    """

    TAG = ""  # No tag
    OFFSET = 0x0  # No offset
    index: int

    def __new__(cls, index: int) -> "Node":
        if 0 > index or index > 2**31:
            raise ValidationError(f"{cls} cannot be initialized with value {index}")

        obj = int.__new__(cls, index + cls.OFFSET)
        obj.index = index
        return obj

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.index})"

    def __add__(self, other: int) -> "Node":
        return self.__class__(self.index + other)

    def serialize(self) -> bytes:
        return self.to_bytes(4, byteorder="big")

    def encode(self) -> str:
        return str(self.index) + self.TAG

    @staticmethod
    def decode(node: str) -> Union["SoftNode", "HardNode"]:
        if len(node) < 1:
            raise ValidationError("Cannot use empty string")

        node_class: Union[Type["SoftNode"], Type["HardNode"]]
        if node[-1] in HARD_NODE_SUFFIXES:
            node_class = HardNode
            node_index = node[:-1]
        else:
            node_class = SoftNode
            node_index = node

        try:
            node_value = int(node_index)
        except ValueError as err:
            raise ValidationError(f"'{node_index}' is not a valid node index.") from err

        return node_class(node_value)


class SoftNode(Node):
    """
    Soft node (unhardened), where value = index .
    """

    TAG = ""  # No tag
    OFFSET = 0x0  # No offset


class HardNode(Node):
    """
    Hard node, where value = index + BIP32_HARDENED_CONSTANT .
    """

    TAG = "H"  # "H" (or "'") means hard node (but use "H" for clarity)
    OFFSET = 0x80000000  # 2**31, BIP32 "Hardening constant"


def derive_child_key(
    parent_key: bytes,
    parent_chain_code: bytes,
    node: Node,
) -> Tuple[bytes, bytes]:
    """
    Compute a derivative key from the parent key.

    From BIP32:

    The function CKDpriv((k_par, c_par), i) → (k_i, c_i) computes a child extended
    private key from the parent extended private key:

    1. Check whether the child is a hardened key (i ≥ 2**31).
       If the child is a hardened key,
       let I = HMAC-SHA512(Key = c_par, Data = 0x00 || ser_256(k_par) || ser_32(i)).
       (Note: The 0x00 pads the private key to make it 33 bytes long.)
       If it is not a hardened key, then
       let I = HMAC-SHA512(Key = c_par, Data = ser_P(point(k_par)) || ser_32(i)).
    2. Split I into two 32-byte sequences, I_L and I_R.
    3. The returned child key k_i is parse_256(I_L) + k_par (mod n).
    4. The returned chain code c_i is I_R.
    5. In case parse_256(I_L) ≥ n or k_i = 0, the resulting key is invalid,
       and one should proceed with the next value for i.
       (Note: this has probability lower than 1 in 2**127.)

    """
    assert len(parent_chain_code) == 32
    if isinstance(node, HardNode):
        # NOTE Empty byte is added to align to SoftNode case
        assert len(parent_key) == 32  # Should be guaranteed here in return statement
        child = hmac_sha512(parent_chain_code, b"\x00" + parent_key + node.serialize())

    elif isinstance(node, SoftNode):
        assert len(ec_point(parent_key)) == 33  # Should be guaranteed by Account class
        child = hmac_sha512(parent_chain_code, ec_point(parent_key) + node.serialize())

    else:
        raise ValidationError(f"Cannot process: {node}")

    assert len(child) == 64

    if to_int(child[:32]) >= SECP256K1_N:
        # Invalid key, compute using next node (< 2**-127 probability)
        return derive_child_key(parent_key, parent_chain_code, node + 1)

    child_key = (to_int(child[:32]) + to_int(parent_key)) % SECP256K1_N
    if child_key == 0:
        # Invalid key, compute using next node (< 2**-127 probability)
        return derive_child_key(parent_key, parent_chain_code, node + 1)

    child_key_bytes = child_key.to_bytes(32, byteorder="big")
    child_chain_code = child[32:]
    return child_key_bytes, child_chain_code


class HDPath:
    def __init__(self, path: str):
        """
        Create a new Hierarchical Deterministic path by decoding the given path.

        Initializes an hd account generator using the
        given path string (from BIP-0032). The path is decoded into nodes of the
        derivation key tree, which define a pathway from a given main seed to
        the child key that is used for a given purpose. Please also reference BIP-
        0043 (which definites the first level as the "purpose" field of an HD path)
        and BIP-0044 (which defines a commonly-used, 5-level scheme for BIP32 paths)
        for examples of how this object may be used. Please note however that this
        object makes no such assumptions of the use of BIP43 or BIP44, or later BIPs.
        :param path             : BIP32-compatible derivation path
        :type path              : str as "m/idx_0/.../idx_n" or "m/idx_0/.../idx_n"
                                  where idx_* is either an integer value (soft node)
                                  or an integer value followed by either the "'" char
                                  or the "H" char (hardened node)
        """
        if len(path) < 1:
            raise ValidationError("Cannot parse path from empty string.")

        nodes = path.split("/")  # Should at least make 1 entry in resulting list
        if nodes[0] not in BASE_NODE_IDENTIFIERS:
            raise ValidationError(f'Path is not valid: "{path}". Must start with "m"')

        decoded_path = []
        for _idx, node in enumerate(nodes[1:]):  # We don't need the root node 'm'
            try:
                decoded_path.append(Node.decode(node))
            except ValidationError as err:
                raise ValidationError(
                    f'Path "{path}" is not valid. Issue with node "{node}": {err}'
                ) from err

        self._path = decoded_path

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(path="{self.encode()}")'

    def encode(self) -> str:
        """
        Encodes this class to a string (reversing the decoding in the constructor).
        """
        encoded_path = ("m",) + tuple(node.encode() for node in self._path)
        return "/".join(encoded_path)

    def derive(self, seed: bytes) -> bytes:
        """
        Perform the BIP32 Hierarchical Derivation recursive loop with the given Path.

        Note that the key and chain_code are initialized with the main seed, and that
        the key that is returned is the child key at the end of derivation process (and
        the chain code is discarded)
        """
        main_node = hmac_sha512(b"Bitcoin seed", seed)
        key = main_node[:32]
        chain_code = main_node[32:]
        for node in self._path:
            key, chain_code = derive_child_key(key, chain_code, node)
        return key



#!/usr/bin/env python
from setuptools import (
    find_packages,
    setup,
)

extras_require = {
    "dev": [
        "build>=0.9.0",
        "bumpversion>=0.5.3",
        "ipython",
        "mypy==1.10.0",
        "pre-commit>=3.4.0",
        "tox>=4.0.0",
        "twine",
        "wheel",
    ],
    "docs": [
        "sphinx>=6.0.0",
        "sphinx-autobuild>=2021.3.14",
        "sphinx_rtd_theme>=1.0.0",
        "towncrier>=21,<22",
    ],
    "test": [
        "pytest>=7.0.0",
        "pytest-xdist>=2.4.0",
        "hypothesis>=4.18.0,<5",
        "coverage",
    ],
}

extras_require["dev"] = (
    extras_require["dev"] + extras_require["docs"] + extras_require["test"]
)


with open("./README.md") as readme:
    long_description = readme.read()


setup(
    name="eth-account",
    # *IMPORTANT*: Don't manually change the version here. Use `make bump`, as described in readme
    version="0.13.3",
    description="""eth-account: Sign Ethereum transactions and messages with local private keys""",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="The Ethereum Foundation",
    author_email="snakecharmers@ethereum.org",
    url="https://github.com/ethereum/eth-account",
    include_package_data=True,
    install_requires=[
        "bitarray>=2.4.0",
        "eth-abi>=4.0.0-b.2",
        "eth-keyfile>=0.7.0",
        "eth-keys>=0.4.0",
        "eth-rlp>=2.1.0",
        "eth-utils>=2.0.0",
        "hexbytes>=1.2.0",
        "rlp>=1.0.0",
        "ckzg>=2.0.0",
        "pydantic>=2.0.0",
    ],
    python_requires=">=3.8, <4",
    extras_require=extras_require,
    py_modules=["eth_account"],
    license="MIT",
    zip_safe=False,
    keywords="ethereum",
    packages=find_packages(exclude=["scripts", "scripts.*", "tests", "tests.*"]),
    package_data={
        "eth_account": [
            "py.typed",
            "hdaccount/wordlist/*.txt",
        ]
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
)



# Originally from: https://github.com/trezor/python-mnemonic
#
# Copyright (c) 2013 Pavol Rusnak
# Copyright (c) 2017 mruddy
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
import pytest

from eth_account.hdaccount.mnemonic import (
    Mnemonic,
    ValidationError,
    normalize_string,
)


def test_failed_checksum():
    mnemo = Mnemonic("english")
    assert not mnemo.is_mnemonic_valid(
        "bless cloud wheel regular tiny venue bird web grief security dignity zoo"
    )


@pytest.mark.parametrize(
    "language,word",
    [
        ("english", "security"),
        ("french", "baguette"),
        ("spanish", "leche"),
        ("italian", "cordata"),
        ("czech", "nitro"),
        ("japanese", "あらためる"),
        ("korean", "강북"),
        (
            "chinese_simplified",
            "也",  # NOTE: This succeeds even though this character is in both
        ),
        (
            "chinese_traditional",
            "滅",  # traditional and simplified Chinese, assumes simplified
        ),
    ],
)
def test_detection(language, word):
    assert language == Mnemonic.detect_language(word)


def test_undetected_language():
    with pytest.raises(ValidationError):
        Mnemonic.detect_language("xxxxxxx")
    with pytest.raises(ValidationError):
        Mnemonic("xxxxxxx")


def test_expand_word():
    m = Mnemonic("english")
    assert "" == m.expand_word("")
    assert " " == m.expand_word(" ")
    assert "access" == m.expand_word("access")  # word in list
    assert "access" == m.expand_word("acce")  # unique prefix expanded to word in list
    assert "acb" == m.expand_word("acb")  # not found at all
    assert "acc" == m.expand_word("acc")  # multi-prefix match
    assert "act" == m.expand_word("act")  # exact three letter match
    assert "action" == m.expand_word("acti")  # unique prefix expanded to word in list


@pytest.mark.parametrize(
    "lang",
    {
        lang
        for lang in Mnemonic.list_languages()
        if lang
        not in (
            # These languages can't support word expansion
            "japanese",
            "korean",
            "chinese_simplified",
            "chinese_traditional",
        )
    },
)
def test_expand(lang):
    m = Mnemonic(lang)
    # Generates a random set of words, so will never be the same set of words
    words = m.generate()
    for word in words.split(" "):  # Space delinates in languages not excluded above
        # BIP39 can support word expansion with as little as 4 characters
        norm_word = normalize_string(word)
        for size in range(4, len(norm_word)):
            assert m.expand(norm_word[: size + 1]) == word


@pytest.mark.parametrize("lang", Mnemonic.list_languages())
@pytest.mark.parametrize("num_words", [12, 15, 18, 21, 24])
def test_generation(lang, num_words):
    m = Mnemonic(lang)
    mnemonic = m.generate(num_words)
    assert m.is_mnemonic_valid(mnemonic)
    # NOTE: Sometimes traditional chinese can return characters that are also
    # valid simplified chinese characters. In that scenario, the detection
    # algorithm will assume simplified.
    if lang == "chinese_traditional":
        assert "chinese" in Mnemonic.detect_language(mnemonic)
    else:
        assert Mnemonic.detect_language(mnemonic) == lang
    assert len(Mnemonic.to_seed(mnemonic)) == 64


@pytest.mark.parametrize(
    "entropy,expected_mnemonic,expected_seed",
    [
        (
            "00000000000000000000000000000000",
            "abandon abandon abandon abandon abandon abandon "
            "abandon abandon abandon abandon abandon about",
            "c55257c360c07c72029aebc1b53c05ed0362ada38ead3e3e9efa3708e5349553"
            "1f09a6987599d18264c1e1c92f2cf141630c7a3c4ab7c81b2f001698e7463b04",
        ),
        (
            "7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f",
            "legal winner thank year wave sausage "
            "worth useful legal winner thank yellow",
            "2e8905819b8723fe2c1d161860e5ee1830318dbf49a83bd451cfb8440c28bd6f"
            "a457fe1296106559a3c80937a1c1069be3a3a5bd381ee6260e8d9739fce1f607",
        ),
        (
            "80808080808080808080808080808080",
            "letter advice cage absurd amount doctor "
            "acoustic avoid letter advice cage above",
            "d71de856f81a8acc65e6fc851a38d4d7ec216fd0796d0a6827a3ad6ed5511a30"
            "fa280f12eb2e47ed2ac03b5c462a0358d18d69fe4f985ec81778c1b370b652a8",
        ),
        (
            "ffffffffffffffffffffffffffffffff",
            "zoo zoo zoo zoo zoo zoo " "zoo zoo zoo zoo zoo wrong",
            "ac27495480225222079d7be181583751e86f571027b0497b5b5d11218e0a8a13"
            "332572917f0f8e5a589620c6f15b11c61dee327651a14c34e18231052e48c069",
        ),
        (
            "000000000000000000000000000000000000000000000000",
            "abandon abandon abandon abandon abandon abandon "
            "abandon abandon abandon abandon abandon abandon "
            "abandon abandon abandon abandon abandon agent",
            "035895f2f481b1b0f01fcf8c289c794660b289981a78f8106447707fdd9666ca"
            "06da5a9a565181599b79f53b844d8a71dd9f439c52a3d7b3e8a79c906ac845fa",
        ),
        (
            "7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f",
            "legal winner thank year wave sausage "
            "worth useful legal winner thank year "
            "wave sausage worth useful legal will",
            "f2b94508732bcbacbcc020faefecfc89feafa6649a5491b8c952cede496c214a"
            "0c7b3c392d168748f2d4a612bada0753b52a1c7ac53c1e93abd5c6320b9e95dd",
        ),
        (
            "808080808080808080808080808080808080808080808080",
            "letter advice cage absurd amount doctor "
            "acoustic avoid letter advice cage absurd "
            "amount doctor acoustic avoid letter always",
            "107d7c02a5aa6f38c58083ff74f04c607c2d2c0ecc55501dadd72d025b751bc2"
            "7fe913ffb796f841c49b1d33b610cf0e91d3aa239027f5e99fe4ce9e5088cd65",
        ),
        (
            "ffffffffffffffffffffffffffffffffffffffffffffffff",
            "zoo zoo zoo zoo zoo zoo "
            "zoo zoo zoo zoo zoo zoo "
            "zoo zoo zoo zoo zoo when",
            "0cd6e5d827bb62eb8fc1e262254223817fd068a74b5b449cc2f667c3f1f985a7"
            "6379b43348d952e2265b4cd129090758b3e3c2c49103b5051aac2eaeb890a528",
        ),
        (
            "0000000000000000000000000000000000000000000000000000000000000000",
            "abandon abandon abandon abandon abandon abandon "
            "abandon abandon abandon abandon abandon abandon "
            "abandon abandon abandon abandon abandon abandon "
            "abandon abandon abandon abandon abandon art",
            "bda85446c68413707090a52022edd26a1c9462295029f2e60cd7c4f2bbd30971"
            "70af7a4d73245cafa9c3cca8d561a7c3de6f5d4a10be8ed2a5e608d68f92fcc8",
        ),
        (
            "7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f",
            "legal winner thank year wave sausage "
            "worth useful legal winner thank year "
            "wave sausage worth useful legal winner "
            "thank year wave sausage worth title",
            "bc09fca1804f7e69da93c2f2028eb238c227f2e9dda30cd63699232578480a40"
            "21b146ad717fbb7e451ce9eb835f43620bf5c514db0f8add49f5d121449d3e87",
        ),
        (
            "8080808080808080808080808080808080808080808080808080808080808080",
            "letter advice cage absurd amount doctor "
            "acoustic avoid letter advice cage absurd "
            "amount doctor acoustic avoid letter advice "
            "cage absurd amount doctor acoustic bless",
            "c0c519bd0e91a2ed54357d9d1ebef6f5af218a153624cf4f2da911a0ed8f7a09"
            "e2ef61af0aca007096df430022f7a2b6fb91661a9589097069720d015e4e982f",
        ),
        (
            "ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff",
            "zoo zoo zoo zoo zoo zoo "
            "zoo zoo zoo zoo zoo zoo "
            "zoo zoo zoo zoo zoo zoo "
            "zoo zoo zoo zoo zoo vote",
            "dd48c104698c30cfe2b6142103248622fb7bb0ff692eebb00089b32d22484e16"
            "13912f0a5b694407be899ffd31ed3992c456cdf60f5d4564b8ba3f05a69890ad",
        ),
        (
            "9e885d952ad362caeb4efe34a8e91bd2",
            "ozone drill grab fiber curtain grace "
            "pudding thank cruise elder eight picnic",
            "274ddc525802f7c828d8ef7ddbcdc5304e87ac3535913611fbbfa986d0c9e547"
            "6c91689f9c8a54fd55bd38606aa6a8595ad213d4c9c9f9aca3fb217069a41028",
        ),
        (
            "6610b25967cdcca9d59875f5cb50b0ea75433311869e930b",
            "gravity machine north sort system female "
            "filter attitude volume fold club stay "
            "feature office ecology stable narrow fog",
            "628c3827a8823298ee685db84f55caa34b5cc195a778e52d45f59bcf75aba68e"
            "4d7590e101dc414bc1bbd5737666fbbef35d1f1903953b66624f910feef245ac",
        ),
        (
            "68a79eaca2324873eacc50cb9c6eca8cc68ea5d936f98787c60c7ebc74e6ce7c",
            "hamster diagram private dutch cause delay "
            "private meat slide toddler razor book "
            "happy fancy gospel tennis maple dilemma "
            "loan word shrug inflict delay length",
            "64c87cde7e12ecf6704ab95bb1408bef047c22db4cc7491c4271d170a1b213d2"
            "0b385bc1588d9c7b38f1b39d415665b8a9030c9ec653d75e65f847d8fc1fc440",
        ),
        (
            "c0ba5a8e914111210f2bd131f3d5e08d",
            "scheme spot photo card baby mountain "
            "device kick cradle pact join borrow",
            "ea725895aaae8d4c1cf682c1bfd2d358d52ed9f0f0591131b559e2724bb234fc"
            "a05aa9c02c57407e04ee9dc3b454aa63fbff483a8b11de949624b9f1831a9612",
        ),
        (
            "6d9be1ee6ebd27a258115aad99b7317b9c8d28b6d76431c3",
            "horn tenant knee talent sponsor spell "
            "gate clip pulse soap slush warm "
            "silver nephew swap uncle crack brave",
            "fd579828af3da1d32544ce4db5c73d53fc8acc4ddb1e3b251a31179cdb71e853"
            "c56d2fcb11aed39898ce6c34b10b5382772db8796e52837b54468aeb312cfc3d",
        ),
        (
            "9f6a2878b2520799a44ef18bc7df394e7061a224d2c33cd015b157d746869863",
            "panda eyebrow bullet gorilla call smoke "
            "muffin taste mesh discover soft ostrich "
            "alcohol speed nation flash devote level "
            "hobby quick inner drive ghost inside",
            "72be8e052fc4919d2adf28d5306b5474b0069df35b02303de8c1729c9538dbb6"
            "fc2d731d5f832193cd9fb6aeecbc469594a70e3dd50811b5067f3b88b28c3e8d",
        ),
        (
            "23db8160a31d3e0dca3688ed941adbf3",
            "cat swing flag economy stadium alone "
            "churn speed unique patch report train",
            "deb5f45449e615feff5640f2e49f933ff51895de3b4381832b3139941c57b592"
            "05a42480c52175b6efcffaa58a2503887c1e8b363a707256bdd2b587b46541f5",
        ),
        (
            "8197a4a47f0425faeaa69deebc05ca29c0a5b5cc76ceacc0",
            "light rule cinnamon wrap drastic word "
            "pride squirrel upgrade then income fatal "
            "apart sustain crack supply proud access",
            "4cbdff1ca2db800fd61cae72a57475fdc6bab03e441fd63f96dabd1f183ef5b7"
            "82925f00105f318309a7e9c3ea6967c7801e46c8a58082674c860a37b93eda02",
        ),
        (
            "066dca1a2bb7e8a1db2832148ce9933eea0f3ac9548d793112d9a95c9407efad",
            "all hour make first leader extend "
            "hole alien behind guard gospel lava "
            "path output census museum junior mass "
            "reopen famous sing advance salt reform",
            "26e975ec644423f4a4c4f4215ef09b4bd7ef924e85d1d17c4cf3f136c2863cf6"
            "df0a475045652c57eb5fb41513ca2a2d67722b77e954b4b3fc11f7590449191d",
        ),
        (
            "f30f8c1da665478f49b001d94c5fc452",
            "vessel ladder alter error federal sibling "
            "chat ability sun glass valve picture",
            "2aaa9242daafcee6aa9d7269f17d4efe271e1b9a529178d7dc139cd18747090b"
            "f9d60295d0ce74309a78852a9caadf0af48aae1c6253839624076224374bc63f",
        ),
        (
            "c10ec20dc3cd9f652c7fac2f1230f7a3c828389a14392f05",
            "scissors invite lock maple supreme raw "
            "rapid void congress muscle digital elegant "
            "little brisk hair mango congress clump",
            "7b4a10be9d98e6cba265566db7f136718e1398c71cb581e1b2f464cac1ceedf4"
            "f3e274dc270003c670ad8d02c4558b2f8e39edea2775c9e232c7cb798b069e88",
        ),
        (
            "f585c11aec520db57dd353c69554b21a89b20fb0650966fa0a9d6f74fd989d8f",
            "void come effort suffer camp survey "
            "warrior heavy shoot primary clutch crush "
            "open amazing screen patrol group space "
            "point ten exist slush involve unfold",
            "01f5bced59dec48e362f2c45b5de68b9fd6c92c6634f44d6d40aab69056506f0"
            "e35524a518034ddc1192e1dacd32c1ed3eaa3c3b131c88ed8e7e54c49a5d0998",
        ),
    ],
)
def test_english_mnemonics(entropy, expected_mnemonic, expected_seed):
    m = Mnemonic("english")
    mnemonic = m.to_mnemonic(bytes.fromhex(entropy))
    assert m.is_mnemonic_valid(mnemonic)
    assert mnemonic == expected_mnemonic

    seed = Mnemonic.to_seed(mnemonic, passphrase="TREZOR")
    assert seed.hex() == expected_seed


@pytest.mark.parametrize(
    "entropy,expected_mnemonic,passphrase,expected_seed",
    [
        (
            "00000000000000000000000000000000",
            "あいこくしん　あいこくしん　あいこくしん　あいこくしん　あいこくしん　あいこくしん　"
            "あいこくしん　あいこくしん　あいこくしん　あいこくしん　あいこくしん　あおぞら",
            "㍍ガバヴァぱばぐゞちぢ十人十色",
            "a262d6fb6122ecf45be09c50492b31f92e9beb7d9a845987a02cefda57a15f9c"
            "467a17872029a9e92299b5cbdf306e3a0ee620245cbd508959b6cb7ca637bd55",
        ),
        (
            "7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f",
            "そつう　れきだい　ほんやく　わかす　りくつ　ばいか　" "ろせん　やちん　そつう　れきだい　ほんやく　わかめ",
            "㍍ガバヴァぱばぐゞちぢ十人十色",
            "aee025cbe6ca256862f889e48110a6a382365142f7d16f2b9545285b3af64e54"
            "2143a577e9c144e101a6bdca18f8d97ec3366ebf5b088b1c1af9bc31346e60d9",
        ),
        (
            "80808080808080808080808080808080",
            "そとづら　あまど　おおう　あこがれる　いくぶん　けいけん　" "あたえる　いよく　そとづら　あまど　おおう　あかちゃん",
            "㍍ガバヴァぱばぐゞちぢ十人十色",
            "e51736736ebdf77eda23fa17e31475fa1d9509c78f1deb6b4aacfbd760a7e2ad"
            "769c714352c95143b5c1241985bcb407df36d64e75dd5a2b78ca5d2ba82a3544",
        ),
        (
            "ffffffffffffffffffffffffffffffff",
            "われる　われる　われる　われる　われる　われる　" "われる　われる　われる　われる　われる　ろんぶん",
            "㍍ガバヴァぱばぐゞちぢ十人十色",
            "4cd2ef49b479af5e1efbbd1e0bdc117f6a29b1010211df4f78e2ed4008286579"
            "3e57949236c43b9fe591ec70e5bb4298b8b71dc4b267bb96ed4ed282c8f7761c",
        ),
        (
            "000000000000000000000000000000000000000000000000",
            "あいこくしん　あいこくしん　あいこくしん　あいこくしん　あいこくしん　あいこくしん　"
            "あいこくしん　あいこくしん　あいこくしん　あいこくしん　あいこくしん　あいこくしん　"
            "あいこくしん　あいこくしん　あいこくしん　あいこくしん　あいこくしん　あらいぐま",
            "㍍ガバヴァぱばぐゞちぢ十人十色",
            "d99e8f1ce2d4288d30b9c815ae981edd923c01aa4ffdc5dee1ab5fe0d4a3e139"
            "66023324d119105aff266dac32e5cd11431eeca23bbd7202ff423f30d6776d69",
        ),
        (
            "7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f",
            "そつう　れきだい　ほんやく　わかす　りくつ　ばいか　"
            "ろせん　やちん　そつう　れきだい　ほんやく　わかす　"
            "りくつ　ばいか　ろせん　やちん　そつう　れいぎ",
            "㍍ガバヴァぱばぐゞちぢ十人十色",
            "eaaf171efa5de4838c758a93d6c86d2677d4ccda4a064a7136344e975f91fe61"
            "340ec8a615464b461d67baaf12b62ab5e742f944c7bd4ab6c341fbafba435716",
        ),
        (
            "808080808080808080808080808080808080808080808080",
            "そとづら　あまど　おおう　あこがれる　いくぶん　けいけん　"
            "あたえる　いよく　そとづら　あまど　おおう　あこがれる　"
            "いくぶん　けいけん　あたえる　いよく　そとづら　いきなり",
            "㍍ガバヴァぱばぐゞちぢ十人十色",
            "aec0f8d3167a10683374c222e6e632f2940c0826587ea0a73ac5d0493b6a6325"
            "90179a6538287641a9fc9df8e6f24e01bf1be548e1f74fd7407ccd72ecebe425",
        ),
        (
            "ffffffffffffffffffffffffffffffffffffffffffffffff",
            "われる　われる　われる　われる　われる　われる　"
            "われる　われる　われる　われる　われる　われる　"
            "われる　われる　われる　われる　われる　りんご",
            "㍍ガバヴァぱばぐゞちぢ十人十色",
            "f0f738128a65b8d1854d68de50ed97ac1831fc3a978c569e415bbcb431a6a671"
            "d4377e3b56abd518daa861676c4da75a19ccb41e00c37d086941e471a4374b95",
        ),
        (
            "0000000000000000000000000000000000000000000000000000000000000000",
            "あいこくしん　あいこくしん　あいこくしん　あいこくしん　あいこくしん　あいこくしん　"
            "あいこくしん　あいこくしん　あいこくしん　あいこくしん　あいこくしん　あいこくしん　"
            "あいこくしん　あいこくしん　あいこくしん　あいこくしん　あいこくしん　あいこくしん　"
            "あいこくしん　あいこくしん　あいこくしん　あいこくしん　あいこくしん　いってい",
            "㍍ガバヴァぱばぐゞちぢ十人十色",
            "23f500eec4a563bf90cfda87b3e590b211b959985c555d17e88f46f7183590cd"
            "5793458b094a4dccc8f05807ec7bd2d19ce269e20568936a751f6f1ec7c14ddd",
        ),
        (
            "7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f",
            "そつう　れきだい　ほんやく　わかす　りくつ　ばいか　"
            "ろせん　やちん　そつう　れきだい　ほんやく　わかす　"
            "りくつ　ばいか　ろせん　やちん　そつう　れきだい　"
            "ほんやく　わかす　りくつ　ばいか　ろせん　まんきつ",
            "㍍ガバヴァぱばぐゞちぢ十人十色",
            "cd354a40aa2e241e8f306b3b752781b70dfd1c69190e510bc1297a9c5738e833"
            "bcdc179e81707d57263fb7564466f73d30bf979725ff783fb3eb4baa86560b05",
        ),
        (
            "8080808080808080808080808080808080808080808080808080808080808080",
            "そとづら　あまど　おおう　あこがれる　いくぶん　けいけん　"
            "あたえる　いよく　そとづら　あまど　おおう　あこがれる　"
            "いくぶん　けいけん　あたえる　いよく　そとづら　あまど　"
            "おおう　あこがれる　いくぶん　けいけん　あたえる　うめる",
            "㍍ガバヴァぱばぐゞちぢ十人十色",
            "6b7cd1b2cdfeeef8615077cadd6a0625f417f287652991c80206dbd82db17bf3"
            "17d5c50a80bd9edd836b39daa1b6973359944c46d3fcc0129198dc7dc5cd0e68",
        ),
        (
            "ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff",
            "われる　われる　われる　われる　われる　われる　"
            "われる　われる　われる　われる　われる　われる　"
            "われる　われる　われる　われる　われる　われる　"
            "われる　われる　われる　われる　われる　らいう",
            "㍍ガバヴァぱばぐゞちぢ十人十色",
            "a44ba7054ac2f9226929d56505a51e13acdaa8a9097923ca07ea465c4c7e294c"
            "038f3f4e7e4b373726ba0057191aced6e48ac8d183f3a11569c426f0de414623",
        ),
        (
            "77c2b00716cec7213839159e404db50d",
            "せまい　うちがわ　あずき　かろう　めずらしい　だんち　" "ますく　おさめる　ていぼう　あたる　すあな　えしゃく",
            "㍍ガバヴァぱばぐゞちぢ十人十色",
            "344cef9efc37d0cb36d89def03d09144dd51167923487eec42c487f742890854"
            "6fa31a3c26b7391a2b3afe7db81b9f8c5007336b58e269ea0bd10749a87e0193",
        ),
        (
            "b63a9c59a6e641f288ebc103017f1da9f8290b3da6bdef7b",
            "ぬすむ　ふっかつ　うどん　こうりつ　しつじ　りょうり　"
            "おたがい　せもたれ　あつめる　いちりゅう　はんしゃ　ごますり　"
            "そんけい　たいちょう　らしんばん　ぶんせき　やすみ　ほいく",
            "㍍ガバヴァぱばぐゞちぢ十人十色",
            "b14e7d35904cb8569af0d6a016cee7066335a21c1c67891b01b83033cadb3e8a"
            "034a726e3909139ecd8b2eb9e9b05245684558f329b38480e262c1d6bc20ecc4",
        ),
        (
            "3e141609b97933b66a060dcddc71fad1d91677db872031e85f4c015c5e7e8982",
            "くのう　てぬぐい　そんかい　すろっと　ちきゅう　ほあん　"
            "とさか　はくしゅ　ひびく　みえる　そざい　てんすう　"
            "たんぴん　くしょう　すいようび　みけん　きさらぎ　げざん　"
            "ふくざつ　あつかう　はやい　くろう　おやゆび　こすう",
            "㍍ガバヴァぱばぐゞちぢ十人十色",
            "32e78dce2aff5db25aa7a4a32b493b5d10b4089923f3320c8b287a77e5124554"
            "43298351beb3f7eb2390c4662a2e566eec5217e1a37467af43b46668d515e41b",
        ),
        (
            "0460ef47585604c5660618db2e6a7e7f",
            "あみもの　いきおい　ふいうち　にげる　ざんしょ　じかん　" "ついか　はたん　ほあん　すんぽう　てちがい　わかめ",
            "㍍ガバヴァぱばぐゞちぢ十人十色",
            "0acf902cd391e30f3f5cb0605d72a4c849342f62bd6a360298c7013d714d7e58"
            "ddf9c7fdf141d0949f17a2c9c37ced1d8cb2edabab97c4199b142c829850154b",
        ),
        (
            "72f60ebac5dd8add8d2a25a797102c3ce21bc029c200076f",
            "すろっと　にくしみ　なやむ　たとえる　へいこう　すくう　"
            "きない　けってい　とくべつ　ねっしん　いたみ　せんせい　"
            "おくりがな　まかい　とくい　けあな　いきおい　そそぐ",
            "㍍ガバヴァぱばぐゞちぢ十人十色",
            "9869e220bec09b6f0c0011f46e1f9032b269f096344028f5006a6e69ea5b0b8a"
            "fabbb6944a23e11ebd021f182dd056d96e4e3657df241ca40babda532d364f73",
        ),
        (
            "2c85efc7f24ee4573d2b81a6ec66cee209b2dcbd09d8eddc51e0215b0b68e416",
            "かほご　きうい　ゆたか　みすえる　もらう　がっこう　"
            "よそう　ずっと　ときどき　したうけ　にんか　はっこう　"
            "つみき　すうじつ　よけい　くげん　もくてき　まわり　"
            "せめる　げざい　にげる　にんたい　たんそく　ほそく",
            "㍍ガバヴァぱばぐゞちぢ十人十色",
            "713b7e70c9fbc18c831bfd1f03302422822c3727a93a5efb9659bec6ad8d6f2c"
            "1b5c8ed8b0b77775feaf606e9d1cc0a84ac416a85514ad59f5541ff5e0382481",
        ),
        (
            "eaebabb2383351fd31d703840b32e9e2",
            "めいえん　さのう　めだつ　すてる　きぬごし　ろんぱ　" "はんこ　まける　たいおう　さかいし　ねんいり　はぶらし",
            "㍍ガバヴァぱばぐゞちぢ十人十色",
            "06e1d5289a97bcc95cb4a6360719131a786aba057d8efd603a547bd254261c2a"
            "97fcd3e8a4e766d5416437e956b388336d36c7ad2dba4ee6796f0249b10ee961",
        ),
        (
            "7ac45cfe7722ee6c7ba84fbc2d5bd61b45cb2fe5eb65aa78",
            "せんぱい　おしえる　ぐんかん　もらう　きあい　きぼう　"
            "やおや　いせえび　のいず　じゅしん　よゆう　きみつ　"
            "さといも　ちんもく　ちわわ　しんせいじ　とめる　はちみつ",
            "㍍ガバヴァぱばぐゞちぢ十人十色",
            "1fef28785d08cbf41d7a20a3a6891043395779ed74503a5652760ee8c24dfe60"
            "972105ee71d5168071a35ab7b5bd2f8831f75488078a90f0926c8e9171b2bc4a",
        ),
        (
            "4fa1a8bc3e6d80ee1316050e862c1812031493212b7ec3f3bb1b08f168cabeef",
            "こころ　いどう　きあつ　そうがんきょう　へいあん　せつりつ　"
            "ごうせい　はいち　いびき　きこく　あんい　おちつく　"
            "きこえる　けんとう　たいこ　すすめる　はっけん　ていど　"
            "はんおん　いんさつ　うなぎ　しねま　れいぼう　みつかる",
            "㍍ガバヴァぱばぐゞちぢ十人十色",
            "43de99b502e152d4c198542624511db3007c8f8f126a30818e856b2d8a20400d"
            "29e7a7e3fdd21f909e23be5e3c8d9aee3a739b0b65041ff0b8637276703f65c2",
        ),
        (
            "18ab19a9f54a9274f03e5209a2ac8a91",
            "うりきれ　さいせい　じゆう　むろん　とどける　ぐうたら　" "はいれつ　ひけつ　いずれ　うちあわせ　おさめる　おたく",
            "㍍ガバヴァぱばぐゞちぢ十人十色",
            "3d711f075ee44d8b535bb4561ad76d7d5350ea0b1f5d2eac054e869ff7963cdc"
            "e9581097a477d697a2a9433a0c6884bea10a2193647677977c9820dd0921cbde",
        ),
        (
            "18a2e1d81b8ecfb2a333adcb0c17a5b9eb76cc5d05db91a4",
            "うりきれ　うねる　せっさたくま　きもち　めんきょ　へいたく　"
            "たまご　ぜっく　びじゅつかん　さんそ　むせる　せいじ　"
            "ねくたい　しはらい　せおう　ねんど　たんまつ　がいけん",
            "㍍ガバヴァぱばぐゞちぢ十人十色",
            "753ec9e333e616e9471482b4b70a18d413241f1e335c65cd7996f32b66cf9554"
            "6612c51dcf12ead6f805f9ee3d965846b894ae99b24204954be80810d292fcdd",
        ),
        (
            "15da872c95a13dd738fbf50e427583ad61f18fd99f628c417a61cf8343c90419",
            "うちゅう　ふそく　ひしょ　がちょう　うけもつ　めいそう　"
            "みかん　そざい　いばる　うけとる　さんま　さこつ　"
            "おうさま　ぱんつ　しひょう　めした　たはつ　いちぶ　"
            "つうじょう　てさぎょう　きつね　みすえる　いりぐち　かめれおん",
            "㍍ガバヴァぱばぐゞちぢ十人十色",
            "346b7321d8c04f6f37b49fdf062a2fddc8e1bf8f1d33171b65074531ec546d1d"
            "3469974beccb1a09263440fc92e1042580a557fdce314e27ee4eabb25fa5e5fe",
        ),
    ],
)
def test_japanese_mnemonics(entropy, expected_mnemonic, passphrase, expected_seed):
    m = Mnemonic("japanese")
    mnemonic = m.to_mnemonic(bytes.fromhex(entropy))
    assert m.is_mnemonic_valid(mnemonic)
    # NOTE For some reason, the strings weren't appearing in normalized form as
    #      copied rrom BIP39 test vectors
    assert normalize_string(mnemonic) == normalize_string(expected_mnemonic)

    seed = Mnemonic.to_seed(mnemonic, passphrase)
    assert seed.hex() == expected_seed

    # Check this because we had to normalize the string for unicode artifacts
    seed = Mnemonic.to_seed(expected_mnemonic, passphrase)
    assert seed.hex() == expected_seed



from typing import (
    Dict,
    Sequence,
    Union,
)

from eth_keys.datatypes import (
    PrivateKey,
)
from eth_typing import (
    HexStr,
)
from hexbytes import (
    HexBytes,
)

Blobs = Sequence[Union[bytes, HexBytes]]
Bytes32 = bytes
PrivateKeyType = Union[Bytes32, int, HexStr, PrivateKey]

AccessList = Sequence[Dict[str, Union[HexStr, Sequence[HexStr]]]]
RLPStructuredAccessList = Sequence[Sequence[Union[HexStr, Sequence[HexStr]]]]

TransactionDictType = Dict[str, Union[AccessList, bytes, HexStr, int]]
